"""Test Class for Base Class Methods in SCRAPI."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import pytest
from google.genai import types as genai_types
from google.api_core import exceptions
from google.cloud.dialogflowcx_v3beta1 import types
from google.oauth2.credentials import Credentials as UserCredentials
from google.oauth2.service_account import (
    Credentials as ServiceAccountCredentials,
)
from google.protobuf import field_mask_pb2, struct_pb2

from dfcx_scrapi.core.scrapi_base import (
    ScrapiBase,
    api_call_counter_decorator,
    handle_api_error,
    retry_api_call,
    should_retry,
)


@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    default_id = "00000000-0000-0000-0000-000000000000"

    global_parent = f"projects/{project_id}/locations/global"
    global_agent_id = f"{global_parent}/agents/my-agent-1234"
    global_datastore_id = f"{global_parent}/dataStores/test-datastore"
    global_flow_id = f"{global_agent_id}/flows/{default_id}"

    non_global_parent = f"projects/{project_id}/locations/us-central1"
    non_global_agent_id = f"{non_global_parent}/agents/my-agent-1234"
    non_global_datastore_id = f"{non_global_parent}/dataStores/test-datastore"

    email = "mock_email@testing.com"
    creds_path = "/Users/path/to/creds/credentials.json"
    creds_dict = {
        "type": "service_account",
        "project_id": project_id,
        "private_key_id": "1234",
        "private_key": "mock_key",
        "client_email": f"mock-account@{project_id}.iam.gserviceaccount.com",
        "client_id": "1234",
        "universe_domain": "googleapis.com",
    }
    global_scopes = [
        "https://www.googleapis.com/auth/cloud-platform",
        "https://www.googleapis.com/auth/dialogflow",
        ]

    mock_signer = MagicMock()
    mock_signer.key_id = "mock_key_id"
    mock_signer.sign.return_value = b"mock_signature"

    creds_object = ServiceAccountCredentials(
        signer=mock_signer,
        token_uri="https://oauth2.googleapis.com/token",
        service_account_email=email,
        project_id=project_id,
        quota_project_id=project_id,
        scopes=[],
    )
    creds_object.token = "mock_token"
    creds_object.refresh = MagicMock()

    adc_creds = UserCredentials(
        token="mock_user_token",
        client_id="mock_client_id",
        client_secret="mock_client_secret",
        quota_project_id=project_id
    )
    adc_creds.refresh = MagicMock()

    return {
        "project_id": project_id,
        "global_scopes": global_scopes,
        "global_parent": global_parent,
        "global_agent_id": global_agent_id,
        "global_flow_id": global_flow_id,
        "global_datastore_id": global_datastore_id,
        "creds_path": creds_path,
        "creds_dict": creds_dict,
        "creds_object": creds_object,
        "adc_creds_object": adc_creds,
        "non_global_agent_id": non_global_agent_id,
        "non_global_datastore_id": non_global_datastore_id,
    }

@pytest.fixture
def mocked_scrapi_base(test_config):
    """Fixture to provide a ScrapiBase instance with mocked default creds."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default:
        mock_adc_creds = test_config["adc_creds_object"]
        mock_default.return_value = (mock_adc_creds, "mock_project_id")
        yield ScrapiBase()


@patch("dfcx_scrapi.core.scrapi_base.default")
def test_init_no_creds(mock_default, test_config):
    """Test initialization with no credentials provided."""
    mock_adc_creds = test_config["adc_creds_object"]
    mock_default.return_value = (mock_adc_creds, "mock_project_id")

    scrapi_base = ScrapiBase()

    assert scrapi_base.creds == mock_adc_creds
    assert isinstance(scrapi_base.creds, UserCredentials)
    assert scrapi_base.token == mock_adc_creds.token
    assert scrapi_base.agent_id is None
    assert not scrapi_base.creds.requires_scopes
    mock_adc_creds.refresh.assert_called_once()
    mock_default.assert_called_once()

def test_init_with_creds(test_config):
    """Test initialization with user provided credentials."""
    mock_creds = test_config["creds_object"]

    scrapi_base = ScrapiBase(creds=mock_creds)

    assert scrapi_base.creds == mock_creds
    assert isinstance(scrapi_base.creds, ServiceAccountCredentials)
    assert scrapi_base.token == mock_creds.token
    assert scrapi_base.agent_id is None
    assert scrapi_base.creds.requires_scopes
    mock_creds.refresh.assert_called_once()


@patch('google.oauth2.service_account.Credentials.from_service_account_file')
def test_init_with_creds_path(mock_from_service_account_file, test_config):
    """Test initialization with credentials path."""
    mock_creds = test_config["creds_object"]

    mock_from_service_account_file.return_value = mock_creds


    scrapi_base = ScrapiBase(creds_path=test_config["creds_path"])

    mock_from_service_account_file.assert_called_once_with(
        test_config["creds_path"], scopes=test_config["global_scopes"])
    assert scrapi_base.creds == mock_creds
    assert isinstance(scrapi_base.creds, ServiceAccountCredentials)
    assert scrapi_base.token == mock_creds.token
    assert scrapi_base.agent_id is None
    assert scrapi_base.creds.requires_scopes
    mock_creds.refresh.assert_called_once()

@patch('google.oauth2.service_account.Credentials.from_service_account_info')
def test_init_with_creds_dict(mock_from_service_account_info, test_config):
    """Test initialization with credentials dictionary."""
    mock_creds = test_config["creds_object"]
    mock_from_service_account_info.return_value = mock_creds

    scrapi_base = ScrapiBase(creds_dict=test_config["creds_dict"])

    mock_from_service_account_info.assert_called_once_with(
        test_config["creds_dict"], scopes=test_config["global_scopes"])
    assert scrapi_base.creds == mock_creds
    assert isinstance(scrapi_base.creds, ServiceAccountCredentials)
    assert scrapi_base.token == mock_creds.token
    assert scrapi_base.agent_id is None
    assert scrapi_base.creds.requires_scopes
    mock_creds.refresh.assert_called_once()

def test_set_region_non_global(test_config):
    """Test _set_region with a non-global location."""
    client_options = ScrapiBase._set_region(test_config["non_global_agent_id"])
    assert client_options["api_endpoint"] == "us-central1-dialogflow.googleapis.com:443"
    assert client_options["quota_project_id"] == test_config["project_id"]

def test_set_region_global(test_config):
    """Test _set_region with a global location."""
    client_options = ScrapiBase._set_region(test_config["global_agent_id"])
    assert client_options["api_endpoint"] == "dialogflow.googleapis.com:443"
    assert client_options["quota_project_id"] == test_config["project_id"]

def test_set_region_invalid_resource_id():
    """Test _set_region with an invalid resource ID."""
    resource_id = "invalid-resource-id"
    with pytest.raises(IndexError):
        ScrapiBase._set_region(resource_id)

def test_client_options_discovery_engine_non_global(test_config):
    """Test _client_options_discovery_engine with a non-global location."""
    client_options = ScrapiBase._client_options_discovery_engine(
        test_config["non_global_datastore_id"])

    assert client_options["api_endpoint"] == "us-central1-discoveryengine.googleapis.com:443"
    assert client_options["quota_project_id"] == test_config["project_id"]

def test_client_options_discovery_engine_global(test_config):
    """Test _client_options_discovery_engine with a global location."""
    client_options = ScrapiBase._client_options_discovery_engine(
        test_config["global_datastore_id"])
  
    assert client_options["api_endpoint"] == "discoveryengine.googleapis.com:443"
    assert client_options["quota_project_id"] == test_config["project_id"]

def test_client_options_discovery_engine_invalid_resource_id():
    """Test _client_options_discovery_engine with an invalid resource ID."""
    resource_id = "invalid-resource-id"
    with pytest.raises(IndexError):
        ScrapiBase._client_options_discovery_engine(resource_id)

def test_pbuf_to_dict():
    """Test pbuf_to_dict."""
    # Create a sample protobuf message
    message = struct_pb2.Struct()
    message["field1"] = "value1"
    message["field2"] = 123

    # Convert to dictionary
    result = ScrapiBase.pbuf_to_dict(message)

    # Assert the result
    assert isinstance(result, dict)
    assert result["field1"] == "value1"
    assert result["field2"] == 123

def test_dict_to_struct():
    """Test dict_to_struct."""
    input_dict = {"field1": "value1", "field2": 123}
    result = ScrapiBase.dict_to_struct(input_dict)
    assert isinstance(result, struct_pb2.Struct)
    assert result["field1"] == "value1"
    assert result["field2"] == 123

def test_parse_agent_id(test_config):
    """Test parse_agent_id with a valid resource ID."""
    resource_id = test_config["global_flow_id"]
    agent_id = ScrapiBase.parse_agent_id(resource_id)
    assert agent_id == test_config["global_agent_id"]

def test_parse_agent_id_short_resource_id(test_config):
    """Test parse_agent_id with a short resource ID."""
    with pytest.raises(ValueError):
        ScrapiBase.parse_agent_id(test_config["global_parent"])

def test_parse_agent_id_invalid_resource_id(test_config):
    """Test parse_agent_id with an invalid resource ID."""
    with pytest.raises(ValueError):
        ScrapiBase.parse_agent_id(test_config["global_datastore_id"])

@patch('dfcx_scrapi.core.scrapi_base.api_call_counter_decorator') 
def test_api_call_counter_decorator(mock_decorator):
    """Test api_call_counter_decorator."""
    mock_decorator.side_effect = lambda func: func # noop

    def mock_api_call():
        pass

    decorated_func = api_call_counter_decorator(mock_api_call)
    assert hasattr(decorated_func, "calls_api")
    assert decorated_func.calls_api is True

def test_should_retry():
    """Test should_retry with different exception types."""
    assert should_retry(exceptions.TooManyRequests("Too many requests")) is True
    assert should_retry(exceptions.ServerError("Server Error")) is True
    assert should_retry(exceptions.BadRequest("Bad Request")) is False
    assert should_retry(ValueError("Value error")) is False

@patch('time.sleep')
def test_retry_api_call_success(mock_sleep):
    """Test retry_api_call with a successful API call."""

    @retry_api_call([1, 2])
    def mock_api_call():
        return "success"

    result = mock_api_call()
    assert result == "success"
    mock_sleep.assert_not_called()

@patch('time.sleep')
def test_retry_api_call_too_many_requests(mock_sleep):
    """Test retry_api_call with TooManyRequests exception."""

    @retry_api_call([1, 2])
    def mock_api_call():
        raise exceptions.TooManyRequests("Too many requests")

    with pytest.raises(exceptions.TooManyRequests):
        mock_api_call()

    mock_sleep.assert_called_with(2)  # Second retry interval

@patch('time.sleep')
def test_retry_api_call_server_error(mock_sleep):
    """Test retry_api_call with ServerError exception."""

    @retry_api_call([1, 2])
    def mock_api_call():
        raise exceptions.ServerError("Server error")

    with pytest.raises(exceptions.ServerError):
        mock_api_call()

    mock_sleep.assert_called_with(2)  # Second retry interval

@patch('time.sleep')
def test_retry_api_call_bad_request(mock_sleep):
    """Test retry_api_call with BadRequest exception."""

    @retry_api_call([1, 2])
    def mock_api_call():
        raise exceptions.BadRequest("Bad request")

    with pytest.raises(exceptions.BadRequest):
        mock_api_call()

    mock_sleep.assert_not_called()  # No retries for BadRequest

def test_handle_api_error_success():
    """Test handle_api_error with a successful API call."""

    @handle_api_error
    def mock_api_call():
        return "success"

    result = mock_api_call()
    assert result == "success"

def test_handle_api_error_google_api_call_error():
    """Test handle_api_error with GoogleAPICallError exception."""

    @handle_api_error
    def mock_api_call():
        raise exceptions.GoogleAPICallError("API error")

    result = mock_api_call()
    assert result is None


def test_handle_api_error_value_error():
    """Test handle_api_error with ValueError exception."""

    @handle_api_error
    def mock_api_call():
        raise ValueError("Value error")

    with pytest.raises(ValueError):
        mock_api_call()

def test_update_kwargs_with_kwargs():
    """Test _update_kwargs with kwargs provided."""
    environment = types.Environment()
    kwargs = {"display_name": "New Display Name", "description": "Updated Description"}
    field_mask = ScrapiBase._update_kwargs(environment, **kwargs)

    assert environment.display_name == "New Display Name"
    assert environment.description == "Updated Description"
    assert field_mask == field_mask_pb2.FieldMask(paths=["display_name", "description"])

def test_update_kwargs_no_kwargs():
    """Test _update_kwargs with no kwargs provided."""
    environment = types.Environment()
    field_mask = ScrapiBase._update_kwargs(environment)

    # Assert that the field mask includes all expected paths for Environment
    expected_paths = [
        "name", "display_name", "description", "version_configs",
        "update_time", "test_cases_config", "webhook_config",
    ]
    assert field_mask == field_mask_pb2.FieldMask(paths=expected_paths)

def test_update_kwargs_experiment():
    """Test _update_kwargs with an Experiment object."""
    experiment = types.Experiment()
    field_mask = ScrapiBase._update_kwargs(experiment)

    # Assert that the field mask includes all expected paths for Experiment
    expected_paths = [
        "name", "display_name", "description", "state", "definition",
        "rollout_config", "rollout_state", "rollout_failure_reason",
        "result", "create_time", "start_time", "end_time",
        "last_update_time", "experiment_length", "variants_history",
    ]
    assert field_mask == field_mask_pb2.FieldMask(paths=expected_paths)

def test_update_kwargs_test_case():
    """Test _update_kwargs with a TestCase object."""
    test_case = types.TestCase()
    field_mask = ScrapiBase._update_kwargs(test_case)

    # Assert that the field mask includes all expected paths for TestCase
    expected_paths = [
        "name", "tags", "display_name", "notes", "test_config",
        "test_case_conversation_turns", "creation_time",
        "last_test_result",
    ]
    assert field_mask == field_mask_pb2.FieldMask(paths=expected_paths)

def test_update_kwargs_version():
    """Test _update_kwargs with a Version object."""
    version = types.Version()
    field_mask = ScrapiBase._update_kwargs(version)

    # Assert that the field mask includes all expected paths for Version
    expected_paths = [
        "name", "display_name", "description", "nlu_settings",
        "create_time", "state",
    ]
    assert field_mask == field_mask_pb2.FieldMask(paths=expected_paths)

def test_update_kwargs_invalid_object():
    """Test _update_kwargs with an invalid object type."""
    with pytest.raises(ValueError) as err:
        ScrapiBase._update_kwargs("invalid_object")

    assert str(err.value) == (
        "`obj` should be one of the following: "
        "[Environment, Experiment, TestCase, Version]."
    )

def test_get_api_calls_details_no_calls(mocked_scrapi_base):
    """Test get_api_calls_details with no API calls made."""
    api_calls_details = mocked_scrapi_base.get_api_calls_details()

    # Assert that the dictionary is empty, as no API calls have been made
    assert api_calls_details == {}

def test_get_api_calls_details_with_calls(mocked_scrapi_base):
    """Test get_api_calls_details with API calls made."""
    @api_call_counter_decorator
    def mock_api_call(self):
        pass

    # Simulate API calls
    mock_api_call(mocked_scrapi_base)  # Call once
    mock_api_call(mocked_scrapi_base)  # Call again

    api_calls_details = mocked_scrapi_base.get_api_calls_details()

    # Assert that the dictionary contains the correct counts
    assert api_calls_details["mock_api_call"] == 2

def test_get_api_calls_count_no_calls(mocked_scrapi_base):
    """Test get_api_calls_count with no API calls made."""
    api_calls_count = mocked_scrapi_base.get_api_calls_count()

    # Assert that the count is 0, as no API calls have been made
    assert api_calls_count == 0

@patch("dfcx_scrapi.core.scrapi_base.default")
def test_get_api_calls_count_with_calls(mocked_scrapi_base):
    """Test get_api_calls_count with API calls made."""
    with patch.object(mocked_scrapi_base, "get_api_calls_count") as mock_count:
        mock_count.return_value = 2

        @api_call_counter_decorator
        def mock_api_call(self):
            pass

        # Simulate API calls
        mock_api_call(mocked_scrapi_base)  # Call once
        mock_api_call(mocked_scrapi_base)  # Call again

        api_calls_count = mocked_scrapi_base.get_api_calls_count()

        # Assert that the count is correct
        assert api_calls_count == 2


def test_get_project_id_from_agent (mocked_scrapi_base):
    project_id = mocked_scrapi_base.get_project_id_from_agent_id(
        resource_id="projects/p1234/locations/global/agents/a1234"
        )

    assert project_id == "p1234"


def test_get_location_id_from_agent(mocked_scrapi_base):
    location_id = mocked_scrapi_base.get_location_id_from_agent_id(
        resource_id="projects/p1234/locations/global/agents/a1234")

    assert location_id == "global"

def test_get_generation_config_invalid_parameters():
    """Test get_generation_config function with invalid parameters.
    This test ensures that the function ignores invalid parameters and
    still creates a valid GenerateContentConfig object.
    """
    parameters = {
        "temperature": 0.9,
        "invalid_parameter": "some_value"
    }

    gen_config = ScrapiBase._get_generate_content_config(parameters)

    assert isinstance(gen_config, genai_types.GenerateContentConfig)

    assert gen_config.temperature == 0.9

    assert not hasattr(gen_config, "invalid_parameter")





"""Search methods for Vertex Search and Conversation."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Any, Dict, List, Optional, Union

from google.cloud.discoveryengine import (
    Document,
    DocumentServiceClient,
    Interval,
    ListDocumentsRequest,
    SearchRequest,
    SearchServiceClient,
    UserInfo,
)

from dfcx_scrapi.core import scrapi_base


class Search(scrapi_base.ScrapiBase):
    """Core Class for Search Client functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

    @staticmethod
    def build_image_query(
        search_request: Dict[str, Any]
    ) -> Union[SearchRequest.ImageQuery, None]:
        image_query = search_request.get("image_query", None)
        if image_query:
            image_bytes = image_query.get("image_bytes", None)
            return SearchRequest.ImageQuery(image_bytes=image_bytes)

        else:
            return None

    @staticmethod
    def build_user_info(
        search_request: Dict[str, Any]
    ) -> Union[UserInfo, None]:

        user_info = search_request.get("user_info", None)
        if user_info:
            user_id = user_info.get("user_id", None)
            user_agent = user_info.get("user_agent", None)
            return UserInfo(user_id=user_id, user_agent=user_agent)

        else:
            return None

    @staticmethod
    def build_interval(interval_dict: Dict[str, Any]) -> Interval:
        for k, v in interval_dict.items():
            if k == "minimum":
                return Interval(minimum=v)
            elif k == "exclusive_minimum":
                return Interval(exclusive_minimum=v)
            elif k == "maximum":
                return Interval(maximum=v)
            elif k == "exclusive_maximum":
                return Interval(exclusive_maximum=v)
            else:
                return None

    @staticmethod
    def search_url(urls: List[str], url: str, regex: bool = False) -> List[str]:
        """Searches a url in a list of urls."""
        matched_urls: List[str] = []

        if regex:
            pattern = re.compile(url)
            for item in urls:
                if pattern.search(item):
                    matched_urls.append(item)
                    print(item)

        else:
            for item in urls:
                if url in item:
                    matched_urls.append(item)
                    print(item)

        return matched_urls

    def build_facet_key(
        self, facet_key_dict: Dict[str, Any]
    ) -> SearchRequest.FacetSpec.FacetKey:
        intervals_list = facet_key_dict.get("intervals", None)
        if intervals_list:
            all_intervals = []
            for interval in intervals_list:
                all_intervals.append(self.build_interval(interval))

        return SearchRequest.FacetSpec.FacetKey(
            key=facet_key_dict.get("key", None),
            intervals=all_intervals,
            restricted_values=facet_key_dict.get("restricted_values", None),
            prefixes=facet_key_dict.get("prefixes", None),
            contains=facet_key_dict.get("contains", None),
            case_insensitive=facet_key_dict.get("case_insensitive", False),
            order_by=facet_key_dict.get("order_by", None),
        )

    def build_single_facet_spec(
        self, spec: Dict[str, Any]
    ) -> SearchRequest.FacetSpec:
        facet_key_dict = spec.get("facet_key", None)
        if not facet_key_dict:
            raise ValueError(
                "`facet_key` is required when providing FacetSpec."
            )
        facet_key = self.build_facet_key(facet_key_dict)

        return SearchRequest.FacetSpec(
            facet_key=facet_key,
            limit=spec.get("limit", None),
            excluded_filter_keys=spec.get("excluded_filter_keys", None),
            enable_dynamic_position=spec.get("enable_dynamic_position", None),
        )

    def build_facet_specs(
        self, search_request: Dict[str, Any]
    ) -> Union[List[SearchRequest.FacetSpec], None]:
        facet_specs = search_request.get("facet_specs", None)
        if facet_specs:
            all_specs = []
            for spec in facet_specs:
                all_specs.append(self.build_single_facet_spec(spec))

            return all_specs

        else:
            return None

    def build_condition_boost_spec(
        self, spec: Dict[str, Any]
    ) -> SearchRequest.BoostSpec.ConditionBoostSpec:
        return SearchRequest.BoostSpec.ConditionBoostSpec(
            condition=spec.get("condition", None), boost=spec.get("boost", None)
        )

    def build_boost_spec(
        self, search_request: Dict[str, Any]
    ) -> Union[SearchRequest.BoostSpec, None]:
        boost_spec_dict = search_request.get("boost_spec", None)
        if boost_spec_dict:
            condition_boost_specs = boost_spec_dict.get(
                "condition_boost_specs", None
            )
            all_boost_specs = []
            for spec in condition_boost_specs:
                all_boost_specs.append(self.build_condition_boost_spec(spec))
            return SearchRequest.BoostSpec(
                condition_boost_specs=all_boost_specs
            )

        else:
            return None

    def get_condition_from_map(
            self, exp_spec_dict: Dict[str, Any]
            ) -> SearchRequest.QueryExpansionSpec.Condition:
        condition_map = {
            "DISABLED": SearchRequest.QueryExpansionSpec.Condition.DISABLED, # pylint: disable=C0301
            "AUTO": SearchRequest.QueryExpansionSpec.Condition.AUTO,
        }

        condition_value = exp_spec_dict.get("condition", "DISABLED")

        return condition_map[condition_value]

    def build_query_expansion_spec(
        self, search_request: Dict[str, Any]
    ) -> Union[SearchRequest.QueryExpansionSpec, None]:
        exp_spec_dict = search_request.get("query_expansion_spec", None)
        if exp_spec_dict:
            condition = self.get_condition_from_map(exp_spec_dict)
            pin_unexpanded_results = exp_spec_dict.get(
                "pin_unexpanded_results", False
            )

            return SearchRequest.QueryExpansionSpec(
                condition=condition,
                pin_unexpanded_results=pin_unexpanded_results,
            )

        else:
            return None

    def get_spell_correct_mode_from_map(
        self, spell_spec_dict: Dict[str, Any]
    ) -> SearchRequest.SpellCorrectionSpec.Mode:
        mode_map = {
            "SUGGESTION_ONLY": SearchRequest.SpellCorrectionSpec.Mode.SUGGESTION_ONLY, # noqa: E501
            "AUTO": SearchRequest.SpellCorrectionSpec.Mode.AUTO,
        }

        mode_value = spell_spec_dict.get("mode", "AUTO")

        return mode_map[mode_value]

    def build_spell_correction_spec(
        self, search_request: Dict[str, Any]
    ) -> Union[SearchRequest.SpellCorrectionSpec, None]:
        spell_spec_dict = search_request.get("spell_correction_spec", None)
        if spell_spec_dict:
            mode = self.get_spell_correct_mode_from_map(spell_spec_dict)
            return SearchRequest.SpellCorrectionSpec(mode=mode)

        else:
            return None

    def build_model_prompt_spec(
        self, content_spec_dict: Dict[str, Any]
    ) -> SearchRequest.ContentSearchSpec.SummarySpec.ModelPromptSpec:
        model_prompt_spec_dict = content_spec_dict.get(
            "model_prompt_spec", None
        )
        if model_prompt_spec_dict:
            return SearchRequest.ContentSearchSpec.SummarySpec.ModelPromptSpec(
                preamble=model_prompt_spec_dict.get("preamble", None)
            )

        else:
            return None

    def build_model_spec(
        self, content_spec_dict: Dict[str, Any]
    ) -> SearchRequest.ContentSearchSpec.SummarySpec.ModelSpec:
        model_spec_dict = content_spec_dict.get("model_spec", None)
        if model_spec_dict:
            return SearchRequest.ContentSearchSpec.SummarySpec.ModelSpec(
                version=model_spec_dict.get("version", "stable")
            )

        else:
            return None

    def build_snippet_spec(
            self) -> SearchRequest.ContentSearchSpec.SnippetSpec:
        return SearchRequest.ContentSearchSpec.SnippetSpec(
            return_snippet=True
        )

    def build_summary_spec(
        self, content_spec_dict: Dict[str, Any]
    ) -> SearchRequest.ContentSearchSpec.SummarySpec:

        model_prompt_spec = self.build_model_prompt_spec(content_spec_dict)
        model_spec = self.build_model_spec(content_spec_dict)

        return SearchRequest.ContentSearchSpec.SummarySpec(
            summary_result_count=content_spec_dict.get(
                "summary_result_count", 10
            ),
            include_citations=content_spec_dict.get("include_citations", False),
            ignore_adversarial_query=content_spec_dict.get(
                "ignore_adversarial_query", False
            ),
            ignore_non_summary_seeking_query=content_spec_dict.get(
                "ignore_non_summary_seeking_query", False
            ),
            model_prompt_spec=model_prompt_spec,
            language_code=content_spec_dict.get("language_code", "en"),
            model_spec=model_spec,
        )

    def build_extractive_content_spec(
        self, content_spec_dict: Dict[str, Any]
    ) -> Union[
        SearchRequest.ContentSearchSpec.ExtractiveContentSpec, None
    ]:

        ext_spec_dict = content_spec_dict.get("extractive_content_spec", None)
        if ext_spec_dict:
            return SearchRequest.ContentSearchSpec.ExtractiveContentSpec(
                max_extractive_answer_count=ext_spec_dict.get(
                    "max_extractive_answer_count", 5
                ),
                max_extractive_segment_count=ext_spec_dict.get(
                    "max_extractive_segment_count", 5
                ),
                return_extractive_segment_score=ext_spec_dict.get(
                    "return_extractive_segment_score", False
                ),
                num_previous_segments=ext_spec_dict.get(
                    "num_previous_segments", 5
                ),
                num_next_segments=ext_spec_dict.get("num_next_segments", 5),
            )
        else:
            return None

    def build_content_search_spec(
        self, search_request: Dict[str, Any]
    ) -> Union[SearchRequest.ContentSearchSpec, None]:
        content_spec_dict = search_request.get("content_search_spec", None)
        if content_spec_dict:
            snippet_spec = self.build_snippet_spec()
            summary_spec = self.build_summary_spec(content_spec_dict)
            extractive_content_spec = self.build_extractive_content_spec(
                content_spec_dict
            )

            return SearchRequest.ContentSearchSpec(
                snippet_spec=snippet_spec,
                summary_spec=summary_spec,
                extractive_content_spec=extractive_content_spec,
            )

        else:
            return None

    def build_embedding_vector(
        self, vector_dict: Dict[str, Any]
    ) -> SearchRequest.EmbeddingSpec.EmbeddingVector:
        return SearchRequest.EmbeddingSpec.EmbeddingVector(
            field_path=vector_dict.get("field_path", None),
            vector=vector_dict.get("vector", None),
        )

    def build_embedding_spec(
        self, search_request: Dict[str, Any]
    ) -> Union[SearchRequest.EmbeddingSpec, None]:
        embedding_vectors_dict = search_request.get("embedding_vectors", None)
        if embedding_vectors_dict:
            vector_list = embedding_vectors_dict.get("embedding_vectors", None)
            all_vectors = []
            for vector_dict in vector_list:
                all_vectors.append(self.build_embedding_vector(vector_dict))
            return SearchRequest.EmbeddingSpec(
                embedding_vectors=all_vectors
            )

        else:
            return None

    def list_documents(
            self, datastore_id: str, page_size: int = 1000) -> List[Document]:
        """List all documents in the provided datastore."""
        client_options = self._client_options_discovery_engine(datastore_id)
        client = DocumentServiceClient(
            credentials=self.creds,
            client_options=client_options
        )

        request = ListDocumentsRequest(
            parent=f"{datastore_id}/branches/default_branch",
            page_size=page_size
        )

        response = client.list_documents(request)

        all_docs: List[Document] = []
        for page in response.pages:
            for doc in page.documents:
                all_docs.append(doc)

        return all_docs

    def list_indexed_urls(
            self, datastore_id: str, docs: Optional[List[Document]] = None
            ) -> List[str]:
        """List all indexed URLs from the provided datastore."""
        if not docs:
            docs = self.list_documents(datastore_id)

        urls: List[str] = [doc.content.uri for doc in docs]

        return urls

    def search_doc_id(
            self,
            document_id: str,
            datastore_id: str = None,
            docs: Optional[List[Document]] = None
            ) -> List[str]:
        if not docs and not datastore_id:
            raise ValueError("Must provide either `docs` or `datastore_id`")

        elif not docs and datastore_id:
            docs = self.list_documents(datastore_id)

        doc_found = False
        for doc in docs:
            if doc.parent_document_id == document_id:
                doc_found = True
                print(doc)
                break

        if not doc_found:
            print(f"Document not found for Doc ID: `{document_id}`")

    def check_datastore_index_status(self, datastore_id: str):
        """Checks the current indexing status of your datastore."""

        PENDING_MESSAGE = "No docs found.\n" \
            "It\'s likely one of two issues:\n" \
            "\t[1] Your data store is not finished indexing.\n" \
            "\t[2] Your data store failed indexing.\n\n" \
            "If you just added your data store, it can take up to 4 hours" \
                " before it will become available."

        SUCCESS_MESSAGE = "Success! 🎉\n" \
            "Your indexing is complete.\n" \
            "Your index contains {DOCS} documents."

        docs = self.list_documents(datastore_id)

        if len(docs) == 0:
            print(PENDING_MESSAGE)
        else:
            print(SUCCESS_MESSAGE.replace("{DOCS}", str(len(docs))))


    def search(self, search_config: Dict[str, Any], total_results: int = 10):
        """Performs a search against an indexed Vertex Data Store.

        Args:
            search_config: A dictionary containing keys that correspond to the
                SearchRequest attributes as defined in: https://cloud.google.com/python/docs/reference/discoveryengine/latest/google.cloud.discoveryengine.SearchRequest

                For complex attributes that require nested fields, you can pass
                in another Dictionary as the value.

                Example: To represent the complex facet_specs config with some
                other simple parameters, you would do the following.

				```py
                search_config = {
                    "facet_specs": [
                        {
                        "facet_key": {
                            "key": "my_key",
                            "intervals": [
                                {
                                "minimum": .5
                                },
                                {
                                "maximum": .95
                                }
                            ],
                        "case_insensitive": True
                        },
                        "limit": 10
                        }
                    ],
                    "page_size": 10,
                    "offset": 2
                    }
			total_results: Total number of results to return for the search. If
				not specified, will default to 10 results. Increasing this to a
				high number can result in long search times.

        Returns:
                A List of SearchResponse objects.
        """
        serving_config = (
            f"{search_config.get('data_store_id', None)}"
            "/servingConfigs/default_serving_config"
        )

        branch_stub = "/".join(serving_config.split("/")[0:8])
        branch = branch_stub + "/branches/0"

        request = SearchRequest(
            serving_config=serving_config,
            branch=branch,
            query=search_config.get("query", None),
            image_query=self.build_image_query(search_config),
            page_size=search_config.get("page_size", 10),
            page_token=search_config.get("page_token", None),
            offset=search_config.get("offset", 0),
            filter=search_config.get("filter", None),
            canonical_filter=search_config.get("canonical_filter", None),
            order_by=search_config.get("order_by", None),
            user_info=self.build_user_info(search_config),
            facet_specs=self.build_facet_specs(search_config),
            boost_spec=self.build_boost_spec(search_config),
            params=search_config.get("params", None),
            query_expansion_spec=self.build_query_expansion_spec(search_config),
            spell_correction_spec=self.build_spell_correction_spec(
                search_config
            ),
            user_pseudo_id=search_config.get("user_pseudo_id", None),
            content_search_spec=self.build_content_search_spec(search_config),
            embedding_spec=self.build_embedding_spec(search_config),
            ranking_expression=search_config.get("ranking_expression", None),
            safe_search=search_config.get("safe_search", False),
            user_labels=search_config.get("user_labels", None),
        )

        client_options = self._client_options_discovery_engine(serving_config)
        client = SearchServiceClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.search(request)

        all_results = []
        for search_result in response:
            if len(all_results) < total_results:
                all_results.append(search_result)
            else:
                break

        return all_results



""" Helper functions for en vs lang"""

from google.cloud.dialogflowcx_v3beta1.services import pages
from google.cloud.dialogflowcx_v3beta1.types import page as gcdc_page

from dfcx_scrapi.core.pages import Pages


class PagesChild(Pages):
    """
    Iterates over the pages object to get the fullfillment details
    """
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

    def get_page(self, page_id,language_code) -> gcdc_page.Page:
        """Get a single CX Page object based on the provided Page ID.

            Args:
            page_id: a properly formatted CX Page ID

            Returns:
            A single CX Page Object
            """
        if not page_id:
            page_id = self.page_id
        request = gcdc_page.GetPageRequest()
        request.name=page_id
        request.language_code = language_code
        client_options = self._set_region(page_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_page(request)

        return response



def get_entry_ff(page=None,page_id=None,language_code='en'):
    """
    Returns entry fullfillments stats
    """
    if not page:
        pagesobj=PagesChild()
        page=pagesobj.get_page(page_id=page_id,language_code=language_code)

    payloadc=0
    textc=0
    for i in page.entry_fulfillment.messages:
        try:
            temp=len(i.payload.items())
            payloadc=payloadc+temp
        except Exception:
            pass
        try:
            temp=len(i.text.text)
            textc=textc+temp
        except Exception:
            pass

    return payloadc,textc

def get_param_ff(param=None,page_id=None,idx=None,language_code='en'):
    """
    Returns params fullfillments stats
    """
    if not param:
        pagesobj=PagesChild()
        page=pagesobj.get_page(page_id=page_id,language_code=language_code)
        param=page.form.parameters[idx]
    payloadc=0
    textc=0
    for message in param.fill_behavior.initial_prompt_fulfillment.messages:
        try:
            temp=len(message.payload.items())
            payloadc=payloadc+temp
        except Exception:
            pass
        try:
            temp=len(message.text.text)
            textc=textc+temp
        except Exception:
            pass

    return payloadc,textc

def get_route_ff(route=None,page_id=None,idx=None,language_code='en'):
    """
    Returns route fullfillments stats
    """
    if not route:
        pagesobj=PagesChild()
        page=pagesobj.get_page(page_id=page_id,language_code=language_code)
        route=page.transition_routes[idx]
    payloadc=0
    textc=0
    for i in route.trigger_fulfillment.messages:
        try:
            temp=len(i.payload.items())
            payloadc=payloadc+temp
        except Exception:
            pass
        try:
            temp=len(i.text.text)
            textc=textc+temp
        except Exception:
            pass


    return payloadc,textc



"""Vertex AI Conversation scraper class."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import re
from typing import Any, Union

import gspread
import pandas as pd
from google.oauth2 import service_account
from tqdm.auto import tqdm

from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.scrapi_base import ScrapiBase, retry_api_call
from dfcx_scrapi.core.sessions import Sessions
from dfcx_scrapi.tools.agent_response import AgentResponse

MAX_RETRIES = 5
INPUT_SCHEMA_REQUIRED_COLUMNS = [
    "conversation_id",
    "turn_index",
    "query",
    "expected_answer",
    "expected_uri",
    "user_metadata",
    "parameters"
]

def load_spreadsheet(
        sheet_url: str, worksheet_name: str, credentials: Any
        ) -> pd.DataFrame:
    """Loads the content of a spreadsheet into pandas DataFrame."""
    sheets_client = gspread.authorize(credentials)
    sheet = sheets_client.open_by_url(sheet_url)
    worksheet = sheet.worksheet(worksheet_name)

    return pd.DataFrame(worksheet.get_all_records())

class DataStoreScraper(ScrapiBase):
    """Vertex AI Conversation scraper class."""

    def _extract_url_part(cls, url, pattern):
        pattern_match = pattern.search(url)
        if not pattern_match:
            raise ValueError(f"Invalid url: {url}")

        return pattern_match.group(1)

    @classmethod
    def from_url(
        cls,
        agent_url: str,
        creds: service_account.Credentials = None,
        language_code: str = "en"):
        match = re.search(
            r'projects/[^/]+/locations/[^/]+/agents/[^/]+', agent_url
            )
        if match:
            agent_id = match.group(0)
        else:
            raise ValueError(f"Invalid url: {agent_url}")

        return cls(
            agent_id=agent_id,
            language_code=language_code,
            creds=creds,
        )

    def __init__(
        self,
        agent_id: str,
        language_code: str = "en",
        creds_path: str = None,
        creds_dict: dict[str, str] = None,
        creds=None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
        )

        self.agent_id = agent_id
        self.language_code = language_code

        self.sessions = Sessions(agent_id=self.agent_id)
        self.agents = Agents(creds=self.creds)

    @classmethod
    def _extract_url_part(cls, url, pattern):
        pattern_match = pattern.search(url)
        if not pattern_match:
            raise ValueError(f"Invalid url: {url}")
        return pattern_match.group(1)

    def validate_queryset(self, queryset: pd.DataFrame) -> None:
        "Validates the queryset and raises exception in case of invalid input."
        # validate input schema
        try:
            queryset[INPUT_SCHEMA_REQUIRED_COLUMNS]
        except KeyError as err:
            raise UserWarning(
                "Ensure your input data contains the following columns:"
                f" {INPUT_SCHEMA_REQUIRED_COLUMNS}"
            ) from err

        # validate if conversationd_id and turn_id is unique identifier
        if not (
            queryset["conversation_id"].astype(str)
            + "_"
            + queryset["turn_index"].astype(str)
        ).is_unique:
            raise UserWarning(
                "Ensure that 'conversation_id' and 'turn_index' are unique "
                "identifiers"
            )

        # validate turn_index
        try:
            queryset["turn_index"].astype(int)
        except ValueError as err:
            raise UserWarning(
                "Ensure that 'turn_index' is set as integer"
            ) from err

        if not queryset["turn_index"].astype(int).gt(0).all():
            raise UserWarning("Ensure that 'turn_index' is in [1, inf)")

    def setup_queryset(self, queryset: pd.DataFrame) -> pd.DataFrame:
        """Various Dataframe validation and cleaning functions."""
        queryset = queryset.rename(
            {column: column.lower() for column in queryset.columns}
        )

        self.validate_queryset(queryset)

        queryset["turn_index"] = queryset["turn_index"].astype(int)
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)

        # adding timestamp and agent display name so they can be used as a multi
        # index
        queryset["scrape_timestamp"] = timestamp.isoformat()
        agent_display_name = self.agents.get_agent(self.agent_id).display_name
        queryset["agent_display_name"] = agent_display_name

        queryset = self._create_session_ids(queryset)

        # if the conversation_id can be converted to int then sorting can be
        # done numerically instead of alphabetically
        try:
            queryset["conversation_id"] = queryset["conversation_id"].astype(
                int
            )
        except ValueError:
            pass

        queryset = queryset.sort_values(
            by=["conversation_id", "turn_index"], ascending=True
        )

        return queryset

    def _create_session_ids(self, queryset: pd.DataFrame) -> pd.DataFrame:
        """Creates a unique session id for each conversation_id."""
        sessions = []
        for conversation_id in queryset["conversation_id"].unique():
            sessions.append(
                {
                    "conversation_id": conversation_id,
                    "session_id": self.sessions.build_session_id(self.agent_id),
                }
            )
        sessions_df = pd.DataFrame(sessions)
        return queryset.merge(sessions_df, on="conversation_id", how="left")

    @retry_api_call([i**2 for i in range(MAX_RETRIES)])
    def scrape_detect_intent(
        self,
        query: str,
        session_id: Union[str, None] = None,
        user_metadata: Union[str, None] = None,
        parameters: Union[str, None] = None
    ) -> AgentResponse:
        if session_id is None:
            session_id = self.sessions.build_session_id(self.agent_id)

        if user_metadata:
            try:
                if isinstance(user_metadata, str):
                    user_metadata = json.loads(user_metadata)
            except ValueError as err:
                raise UserWarning("Invalid user metadata") from err

        if parameters:
            try:
                if isinstance(parameters, str):
                    parameters = json.loads(parameters)
            except ValueError as err:
                raise UserWarning("Invalid parameters") from err

        response = self.sessions.detect_intent(
            agent_id=self.agent_id,
            session_id=session_id,
            text=query,
            language_code=self.language_code,
            end_user_metadata=user_metadata,
            populate_data_store_connection_signals=True,
            parameters=parameters
        )

        ar = AgentResponse()
        ar.from_query_result(response._pb)

        return ar

    def run(
        self, queryset: pd.DataFrame, flatten_response: bool = True
    ) -> pd.DataFrame:
        "Runs through each query and concatenates responses to the queryset."
        queryset = self.setup_queryset(queryset)
        progress_bar = tqdm(desc="Scraping queries", total=len(queryset))

        def scrape(row):
            result = self.scrape_detect_intent(
                query=row["query"],
                session_id=row["session_id"],
                user_metadata=row["user_metadata"],
                parameters=row["parameters"]
            )
            progress_bar.update()

            return result

        queryset["query_result"] = queryset.apply(scrape, axis=1)

        return queryset



""" UAT Deployment functions"""

import json
import logging
import sys

from shared.deployment import Deployment

#from .shared.deployments import Deployment
# logging config
logging.basicConfig(
    level=logging.INFO,
    format="UAT: %(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def main(data):
    """
    Deploys and tests a Dialogflow CX agent in a UAT environment.

    This function performs the following steps:

    1. Initializes a Deployment object with the provided data.
    2. Imports the agent to the specified UAT webhook environment.
    3. Validates test cases.
    4. Collects flow IDs.
    5. Deletes versions based on count.
    6. Cuts a new version.
    7. Deploys the new version.
    8. Updates the datastore with UAT information.

    Args:
        data: A dictionary containing configuration data, including the
            'uat_webhook_env' key.
    """

    dep=Deployment(data)
    # call the steps sequentially
    dep.import_agent(webhookenv=data["uat_webhook_env"])
    dep.test_case_validation()
    dep.collect_flow_id()
    dep.version_count_delete()
    dep.version_cut()
    dep.deploy_versions()
    dep.datastore_update("uat")



if __name__=="__main__":
    # read env variables
    with open("config.json" , encoding='utf-8') as config_file:
        config = json.load(config_file)
    SHA_ID=sys.argv[1]
    obj=f"UAT/{config['agent_name']}/{SHA_ID}"
    sha_gs_loc=(
        f"gs://{config['bucket']}/UAT/{config['agent_name']}/{SHA_ID}"
    )
    logging.info("Agent location: %s" ,sha_gs_loc)
    #adding additional variables to dict
    config["sha_agent_gcs_location"]=sha_gs_loc
    config["target_project_id"] = config["uat_project"]
    config["target_environment_name"]=config["uat_env_deploy"]
    with open("agent_artifacts/metadata.json",
              encoding='utf-8') as metadata_file:
        metadata = json.load(metadata_file)

    config["source_flow_names"]=metadata["source_flow_names"]
    config["updated_commit_message"]=metadata["updated_commit_message"]

    # To execute steps in order
    main(config)



"""Unit Tests for Conversation History."""

# pylint: disable=redefined-outer-name
# pylint: disable=protected-access
# pylint: disable=unused-argument

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from unittest.mock import MagicMock, patch

import pytest
from google.cloud.dialogflowcx_v3beta1 import services, types
from google.protobuf import timestamp_pb2

from dfcx_scrapi.core.conversation_history import ConversationHistory


@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    conversation_id = f"{agent_id}/conversations/1234"
    return {
        "project_id": project_id,
        "agent_id": agent_id,
        "conversation_id": conversation_id,
    }

@pytest.fixture
def test_conversation(test_config):
    conversation = types.Conversation(
        name=f"{test_config['conversation_id']}",
        start_time=timestamp_pb2.Timestamp(seconds=1678886400)
        )
    interaction_1 = types.Conversation.Interaction(
        request=types.DetectIntentRequest(
            query_input=types.QueryInput(
                text=types.TextInput(text="Hello"))),
        response=types.DetectIntentResponse(
            query_result=types.QueryResult(
                response_messages=[
                    types.ResponseMessage(
                        text=types.ResponseMessage.Text(
                            text=["Hello there!"]))])))
    interaction_2 = types.Conversation.Interaction(
        request=types.DetectIntentRequest(
            query_input=types.QueryInput(
                text=types.TextInput(text="How are you?"))),
        response=types.DetectIntentResponse(
            query_result=types.QueryResult(
                response_messages=[
                    types.ResponseMessage(
                        text=types.ResponseMessage.Text(
                            text=["I'm doing well, thanks for asking."]
                            ))])))

    conversation.interactions.extend([interaction_1, interaction_2])

    return conversation

@pytest.fixture
def mock_list_conversations_pager(test_conversation):
    return services.conversation_history.pagers\
        .ListConversationsPager(
        services.conversation_history.ConversationHistoryClient\
            .list_conversations,
        types.conversation_history.ListConversationsRequest(),
        types.conversation_history.ListConversationsResponse(
            conversations=[test_conversation]
        ),
    )

@pytest.fixture(autouse=True)
def mock_client(test_config):
    """Setup mock client for all tests."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.conversation_history.services.conversation_history.ConversationHistoryClient") as mock_client:
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client # Return control to test method

# Test get_user_input
def test_get_user_input():
    query_input = types.QueryInput(text=types.TextInput(text="test input"))
    user_input = ConversationHistory.get_user_input(query_input)
    assert user_input == "test input"

# Test get_query_result
def test_get_query_result():
    query_result = types.QueryResult(response_messages=[
        types.ResponseMessage(
            text=types.ResponseMessage.Text(text=["test result"]))
    ])
    agent_response = ConversationHistory.get_query_result(query_result)
    assert agent_response == "test result"

# Test list_conversations
def test_list_conversations(
    mock_client, mock_list_conversations_pager, test_config
    ):
    mock_client.return_value.list_conversations.return_value = \
        mock_list_conversations_pager
    ch = ConversationHistory(agent_id=test_config["agent_id"])
    res = ch.list_conversations(agent_id=test_config["agent_id"])

    assert isinstance(res, list)
    assert isinstance(res[0], types.Conversation)

# Test get_conversation
def test_get_conversation(
    mock_client, test_conversation, test_config
    ):
    mock_client.return_value.get_conversation.return_value = \
        test_conversation
    ch = ConversationHistory(agent_id=test_config["agent_id"])
    res = ch.get_conversation(
        conversation_id=test_config["conversation_id"]
    )

    assert isinstance(res, types.Conversation)

# Test delete_conversation
def test_delete_conversation(mock_client, test_config):
    ch = ConversationHistory(agent_id=test_config["agent_id"])
    ch.delete_conversation(
        conversation_id=test_config["conversation_id"]
    )
    mock_client.return_value.delete_conversation.assert_called()

# Test process_single_conversation
def test_process_single_conversation(test_conversation):
    ch = ConversationHistory()
    conversation = ch.process_single_conversation(test_conversation)
    assert isinstance(conversation, dict)
    assert "session_id" in conversation
    assert "create_time" in conversation
    assert "turns" in conversation
    assert isinstance(conversation["turns"], list)

# Test write_conversations_to_file
def test_write_conversations_to_file(tmpdir):
    ch = ConversationHistory()
    conversations = [{"test": "data"}]
    filename = os.path.join(tmpdir, "test.json")
    ch.write_conversations_to_file(conversations, filename)
    assert os.path.exists(filename)

# Test read_conversations_from_file
def test_read_conversations_from_file(tmpdir):
    ch = ConversationHistory()
    data = [{"test": "data"}]
    filename = os.path.join(tmpdir, "test.json")
    with open(filename, "w", encoding="utf-8") as f:
        for item in data:
            json.dump(item, f)
            f.write("\n")
    loaded_data = ch.read_conversations_from_file(filename)
    assert loaded_data == data

# Test conversation_history_to_file
@patch("dfcx_scrapi.core.conversation_history.thread_map")
def test_conversation_history_to_file(
    mock_thread_map, mock_client, test_conversation, tmpdir, test_config
    ):
    mock_client.return_value.list_conversations.return_value = [
        test_conversation
    ]
    mock_client.return_value.get_conversation.return_value = \
        test_conversation
    mock_thread_map.return_value = [{"test": "data"}]
    agent_id = test_config["agent_id"]
    ch = ConversationHistory()
    filename = os.path.join(tmpdir, "test.json")
    ch.conversation_history_to_file(agent_id, filename)
    assert os.path.exists(filename)



"""Generators Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from typing import Any, Dict, List

from google.cloud.dialogflowcx_v3beta1 import services, types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Generators(scrapi_base.ScrapiBase):
    """Core Class for CX Generators Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if agent_id:
            self.agent_id = agent_id

    @staticmethod
    def __get_placeholders(
        prompt: str,
    ) -> List[types.generator.Generator.Placeholder]:
        """Parse any placeholders from the prompt."""
        placeholders = []
        pattern = r"\$(?:\s+)?([a-zA-Z0-9_]+)"
        matches = re.findall(pattern, prompt)
        for match in matches:
            placeholder = types.generator.Generator.Placeholder()
            placeholder.id = match
            placeholder.name = match
            placeholders.append(placeholder)

        return placeholders

    @staticmethod
    def __input_validation(
        display_name: str, prompt: str, obj: types.Generator
    ):
        """Validate incoming input for Create Generator."""
        if not display_name or obj:
            raise ValueError(
                "At least display_name or obj should be specified."
            )

        if display_name and not obj and not prompt:
            raise ValueError("`prompt` value should be specified.")

    @staticmethod
    def __clean_update_kwargs(
        kwargs: Dict[str, Any], generator: types.Generator
    ):
        """Clean up kwargs before adding to paths for updating the Generator.

        We're providing a few quality of life additions "under the hood" so the
        user doesn't have to recall all of the specific kwargs for the object.
        Specifically, they can pass `prompt` instead of `prompt_text` and we'll
        update appropriately. Additionally we'll detect any placeholders and
        make these updates as well.
        """
        if "prompt" in kwargs:
            kwargs["prompt_text"] = kwargs["prompt"]
            del kwargs["prompt"]

        if len(generator.placeholders) > 0:
            kwargs["placeholders"] = generator.placeholders

        return kwargs

    def get_generators_map(
        self, agent_id: str, reverse=False
    ) -> Dict[str, str]:
        """Put Generator Display Names and UUIDs into a user friendly map."""
        if reverse:
            gen_dict = {
                gen.display_name: gen.name
                for gen in self.list_generators(agent_id=agent_id)
            }

        else:
            gen_dict = {
                gen.name: gen.display_name
                for gen in self.list_generators(agent_id=agent_id)
            }

        return gen_dict

    @scrapi_base.api_call_counter_decorator
    def list_generators(self, agent_id: str) -> List[types.generator.Generator]:
        """Retrieves a list of Generators based on the Agent ID."""

        request = types.generator.ListGeneratorsRequest()
        request.parent = agent_id

        client_options = self._set_region(agent_id)
        client = services.generators.GeneratorsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.list_generators(request)

        generators = []
        for page in response.pages:
            for generator in page.generators:
                generators.append(generator)

        return generators

    @scrapi_base.api_call_counter_decorator
    def get_generator(self, generator_id: str) -> types.Generator:
        """Retrieves a single CX Generator resource object.

        Args:
          generator_id: The formatted CX Generator ID

        Returns:
          A single CX Generator resource object
        """

        request = types.generator.GetGeneratorRequest()
        request.name = generator_id

        client_options = self._set_region(generator_id)
        client = services.generators.GeneratorsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_generator(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_generator(
        self,
        agent_id: str,
        display_name: str = None,
        prompt: str = None,
        obj: types.Generator = None,
    ):
        """Create a Dialogflow CX Generator."""
        self.__input_validation(display_name, prompt, obj)

        if obj:
            generator_obj = obj
            generator_obj.name = ""

        else:
            generator_obj = types.Generator()
            generator_obj.display_name = display_name
            generator_obj.prompt_text.text = prompt
            generator_obj.placeholders = self.__get_placeholders(prompt)

        client_options = self._set_region(agent_id)
        client = services.generators.GeneratorsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.create_generator(
            parent=agent_id, generator=generator_obj
        )

        return response

    @scrapi_base.api_call_counter_decorator
    def update_generator(
        self, generator_id: str, obj: types.Generator = None, **kwargs
    ) -> types.Generator:
        """Updates a single Generator object based on provided kwargs.

        Args:
          generator_id: CX Generator ID string in the following format
              projects/<PROJECT ID>/locations/<LOCATION ID>/
                generators/<GENERATOR ID>
          obj: (Optional) The CX Generator object in proper format. This can
            also be extracted by using the get_generator() method.
          kwargs: You may find a list of generator attributes here:
              https://cloud.google.com/python/docs/reference/dialogflow-cx/
                  latest/google.cloud.dialogflowcx_v3beta1.types.Generator
        Returns:
          The updated CX Generator resource object.
        """

        if obj:
            generator = obj
            generator.name = generator_id
        else:
            generator = self.get_generator(generator_id)

        # set generator attributes to args
        for key, value in kwargs.items():
            if key in ["prompt", "prompt_text"]:
                generator.prompt_text.text = value
                generator.placeholders = self.__get_placeholders(value)
            else:
                setattr(generator, key, value)

        kwargs = self.__clean_update_kwargs(kwargs, generator)
        paths = kwargs.keys()

        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(generator_id)
        client = services.generators.GeneratorsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.update_generator(
            generator=generator, update_mask=mask
        )

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_generator(self, generator_id: str):
        """Deletes the specified Dialogflow CX Generator.

        Args:
          generator_id: CX Generator ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/
              generators/<GENERATOR ID>
        """
        client_options = self._set_region(generator_id)
        client = services.generators.GeneratorsClient(
            credentials=self.creds, client_options=client_options
        )
        client.delete_generator(name=generator_id)



"""CX Session Resource functions."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
from typing import Any, Dict, List

from google.cloud.dialogflowcx_v3beta1 import services, types
from google.protobuf.json_format import MessageToDict
from IPython.display import Markdown, display
from proto.marshal.collections import maps

from dfcx_scrapi.core.environments import Environments
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.playbooks import Playbooks
from dfcx_scrapi.core.scrapi_base import ScrapiBase
from dfcx_scrapi.core.tools import Tools

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Sessions(ScrapiBase):
    """Core Class for CX Session Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
        session_id: str = None,
        tools_map: Dict[str, str] = None,
        playbooks_map: Dict[str, str] = None,
        flows_map: Dict[str, str] = None
    ):
        super().__init__(
            creds_path=creds_path, creds_dict=creds_dict,
            creds=creds, scope=scope
        )

        self._session_id = session_id
        self.agent_id = agent_id
        self.tools_map = tools_map
        self.playbooks_map = playbooks_map
        self.flows_map = flows_map
        self._env_client = None
        self._tools_client = None
        self._playbooks_client = None
        self._flows_client = None

    @property
    def session_id(self):
        """Property for the session ID, parses the resource path if needed."""
        if self._session_id:
            self._parse_resource_path("session", self._session_id)

        return self._session_id

    @property
    def playbooks_client(self):
        """Property for Playbooks client."""
        if self._playbooks_client is None:
            if not self.agent_id:
                raise ValueError(
                    "agent_id must be set to use Playbooks Client.")
            self._playbooks_client = Playbooks(
                agent_id=self.agent_id, creds=self.creds
            )
        return self._playbooks_client

    @property
    def tools_client(self):
        """Property for Tools client."""
        if self._tools_client is None:
            self._tools_client = Tools(creds=self.creds)

        return self._tools_client

    @property
    def flows_client(self):
        """Property for Flows client."""
        if self._flows_client is None:
            self._flows_client = Flows(creds=self.creds)

        return self._flows_client

    @property
    def env_client(self):
        """Property for Environments client."""
        if self._env_client is None:
            self._env_client = Environments(creds=self.creds)

        return self._env_client

    @staticmethod
    def printmd(string):
        display(Markdown(string))

    @staticmethod
    def _build_query_input(text, language_code):
        """Build out the query_input object for the Query Request.

        Args:
          text, the text to use for the Detect Intent request.
          language_code, the language code to use for Detect Intent request.
        """
        text_input = types.session.TextInput(text=text)
        query_input = types.session.QueryInput(
            text=text_input, language_code=language_code
        )

        return query_input

    @staticmethod
    def build_intent_query_input(intent_id: str, language_code: str):
        """Build the query_input object for direct Intent request."""
        intent_input = types.session.IntentInput(intent=intent_id)
        query_input = types.session.QueryInput(
            intent=intent_input, language_code=language_code
        )

        return query_input

    @staticmethod
    def get_tool_action(tool_use: types.trace.ToolUse) -> str:
        return tool_use.action

    def get_tool_params(self, params: maps.MapComposite):
        "Handle various types of param values from Tool input/outputs."
        param_map = {}
        if isinstance(params, maps.MapComposite):
            param_map = self.recurse_proto_marshal_to_dict(params)

        # Clean up resulting param map. This is because I/O params from Agent
        # Builder proto will have a blank top level key, but the main value
        # info is what is important for return to the user in this tool.
        empty_top_key = param_map.get("", None)
        if len(param_map.keys()) == 1 and empty_top_key:
            param_map = param_map[""]

        return param_map

    def get_playbook_name(self, playbook_id: str):
        agent_id = self.parse_agent_id(playbook_id)
        if not self.playbooks_map:
            self.playbooks_map = self.playbooks_client.get_playbooks_map(
                agent_id)

        return self.playbooks_map[playbook_id]

    def get_tool_name(self, tool_use: types.trace.ToolUse) -> str:
        agent_id = self.parse_agent_id(tool_use.tool)
        if not self.tools_map:
            self.tools_map = self.tools_client.get_tools_map(agent_id)

        return self.tools_map[tool_use.tool]

    def get_flow_name(self, flow_id: str):
        agent_id = self.parse_agent_id(flow_id)
        if not self.flows_map:
            self.flows_map = self.flows_client.get_flows_map(agent_id)

        return self.flows_map[flow_id]

    def collect_tool_responses(
        self, res: types.session.QueryResult
    ) -> List[Dict[str, str]]:
        """Gather all the tool responses into a list of dicts."""
        tool_responses = []
        for action in res.generative_info.action_tracing_info.actions:
            if action.tool_use:
                tool_responses.append(
                    {
                        "tool_name": self.get_tool_name(action.tool_use),
                        "tool_action": self.get_tool_action(action.tool_use),
                        "input_params": self.get_tool_params(
                            action.tool_use.input_action_parameters),
                        "output_params": self.get_tool_params(
                            action.tool_use.output_action_parameters),
                    }
                )

        return tool_responses

    def collect_playbook_responses(
        self, res: types.session.QueryResult
    ) -> List[Dict[str, str]]:
        """Gather all the playbook responses into a list of dicts."""
        playbook_responses = []
        for action in res.generative_info.action_tracing_info.actions:
            if action.playbook_invocation:
                playbook_responses.append(
                    {
                        "playbook_name": self.get_playbook_name(
                            action.playbook_invocation.playbook
                        )
                    }
                )
            # If no playbook invocation was found try to return the current
            # Playbook
            elif len(res.generative_info.current_playbooks) > 0:
                playbook_responses.append(
                    {
                        "playbook_name": self.get_playbook_name(
                            res.generative_info.current_playbooks[-1]
                            )
                    }
                )

        return playbook_responses

    def collect_flow_responses(
        self, res: types.session.QueryResult
    ) -> List[Dict[str, str]]:
        """Gather all the flow repsonses into a list of dicts."""
        flow_responses = []
        for action in res.generative_info.action_tracing_info.actions:
            if action.flow_invocation:
                flow_responses.append(
                    {
                        "flow_name": self.get_flow_name(
                            action.flow_invocation.flow
                        )
                    }
                )

        return flow_responses

    def build_session_id(
        self, agent_id: str = None, overwrite: bool = True,
        environment_name: str = None
    ) -> str:
        """Creates a valid UUID-4 Session ID to use with other methods.

        Args:
          agent_id: the Agent ID of the CX Agent.
          overwrite (Optional), if a session_id already exists, this will
            overwrite the existing Session ID parameter. Defaults to True.
          environment_name: (Optional) the human readable Environment name to
            use when building the session ID. If this is not provided, DRAFT is
            assumed.
        """

        # Parse and validate the incoming agent_id
        _ = self._parse_resource_path("agent", agent_id)


        if environment_name:
            env = self.env_client.get_environment_by_display_name(
                environment_name, agent_id
            )
            if not env:
                raise ValueError(f"Environment `{environment_name}` does not"
                                 " exist.")
            session_id = f"{env.name}/sessions/{uuid.uuid4()}"

        else:
            session_id = f"{agent_id}/sessions/{uuid.uuid4()}"

        if overwrite:
            self._session_id = session_id

        return session_id

    def detect_intent(
        self,
        agent_id,
        session_id,
        text: str = None,
        language_code: str = "en",
        parameters: Dict[str, Any] = None,
        end_user_metadata: Dict[str, Any] = None,
        populate_data_store_connection_signals: bool = False,
        intent_id: str = None,
        timezone: str = None
    ):
        """Returns the result of detect intent with texts as inputs.

        Using the same `session_id` between requests allows continuation
        of the conversation.

        Args:
          agent_id: the Agent ID of the CX Agent to have the conversation with.
          session_id: an RFC 4122 formatted UUID to be used as the unique ID
            for the duration of the conversation session. When using Python
            uuid library, uuid.uuid4() is preferred.
          text: (Optional) the user utterance to run intent detection on
          language_code: (Optional) corresponds to the language code to use with
            query inputs to the agent.
          parameters: (Optional) Dict of CX Session Parameters to set in the
            conversation. Typically this is set before a conversation starts.
          end_user_metadata: (Optional) Dict of CX Session endUserMetadata to
            set in the conversation.
          populate_data_store_connection_signals: If set to true and data
            stores are involved in serving the request then query result will
            be populated with data_store_connection_signals field which
            contains data that can help evaluations.
          intent_id: fully qualified Intent ID path to pass in for query
            input instead of text. This allows for the direct triggering of a
            specific Intent, and will bypass the NLU engine.
          timezone: (Optional) IANA Timezone database code to pass in with query
            input   which can be used by the agent runtime. For example, when
            capturing datetime via system functions, they can be modified to
            user the provied timezone vs. the default agent timezone.
            Refs: https://www.iana.org/time-zones

        Returns:
          The CX query result from intent detection
        """
        client_options = self._set_region(agent_id)
        session_client = services.sessions.SessionsClient(
            client_options=client_options, credentials=self.creds
        )

        res = self._parse_resource_path("session", str(session_id), False)
        if not res:
            raise ValueError(
                "Session ID must be provided in the following format: "
                "`projects/<Project ID>/locations/<Location ID>/agents/"
                "<Agent ID>/sessions/<Session ID>`.\n\n"
                "Utilize `build_session_id` to create a new Session ID."
            )

        if intent_id:
            query_input = self.build_intent_query_input(
                intent_id, language_code)
        else:
            query_input = self._build_query_input(text, language_code)

        request = types.session.DetectIntentRequest()
        request.session = session_id
        request.query_input = query_input

        query_param_mapping = {}

        if parameters:
            query_param_mapping["parameters"] = parameters

        if end_user_metadata:
            query_param_mapping["end_user_metadata"] = end_user_metadata

        if populate_data_store_connection_signals:
            query_param_mapping[
                "populate_data_store_connection_signals"
            ] = populate_data_store_connection_signals

        if timezone:
            query_param_mapping["time_zone"] = timezone

        if query_param_mapping:
            query_params = types.session.QueryParameters(query_param_mapping)
            request.query_params = query_params

        response = session_client.detect_intent(request=request)
        query_result = response.query_result

        return query_result

    def get_agent_answer(self, user_query: str) -> str:
        """Extract the answer/citation from a Vertex Conversation response."""

        session_id = self.build_session_id(self.agent_id)
        res = MessageToDict(
            self.detect_intent(  # pylint: disable=W0212
                self.agent_id, session_id, user_query
            )._pb
        )

        answer_text = res["responseMessages"][0]["text"]["text"][0]
        answer_link = (
            res["responseMessages"][1]["payload"]["richContent"][0][0][
                "actionLink"
            ]
            if len(res["responseMessages"]) > 1
            else ""
        )

        return f"{answer_text} ({answer_link})"

    def parse_result(self, res):

        tool_call_font = "<font color='dark red'>TOOL CALL:</font></b>"
        tool_res_font = "<font color='yellow'>TOOL RESULT:</font></b>"
        query_font = "<font color='green'><b> USER QUERY:</font></b>"
        response_font = "<font color='green'><b>AGENT RESPONSE:</font></b>"

        self.printmd(f"{query_font} {res.text}")

        for action in res.generative_info.action_tracing_info.actions:

            if "tool_use" in action:
                tool_name = action.tool_use.action
                input_params = {}
                output_params = {}

                input_param = action.tool_use.input_action_parameters
                output_param = action.tool_use.output_action_parameters

                if isinstance(input_param, maps.MapComposite):
                    processed_input_params = self.recurse_proto_marshal_to_dict(
                        input_param
                    )
                    input_keys = list(processed_input_params.keys())
                    first_key = input_keys[0] if input_keys else None
                    input_params = processed_input_params.get(first_key, {})

                if isinstance(output_param, maps.MapComposite):
                    processed_output_params = (
                        self.recurse_proto_marshal_to_dict(output_param)
                    )
                    output_keys = list(processed_output_params.keys())
                    first_key = output_keys[0] if output_keys else None

                    output_params = processed_output_params.get(first_key, {})

                self.printmd(f"{tool_call_font} {tool_name} -> {input_params}")
                self.printmd(f"{tool_res_font} {output_params}")

            elif "agent_utterance" in action:
                self.printmd(f"{response_font} {action.agent_utterance.text}")



"""Test Class for Test Cases in SCRAPI."""

# pylint: disable=redefined-outer-name
# pylint: disable=protected-access

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
from google.cloud.dialogflowcx_v3beta1 import types
from google.cloud.dialogflowcx_v3beta1.services import test_cases

from dfcx_scrapi.core.test_cases import TestCases as PyTestCases


@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    flow_id = f"{agent_id}/flows/00000000-0000-0000-0000-000000000000"
    page_id = f"{flow_id}/pages/mock-page-1234"
    other_flow_id = f"{agent_id}/flows/other1234"
    other_page_id = f"{other_flow_id}/pages/other1234"
    test_case_id = f"{agent_id}/testCases/1234"
    test_case_result_id = f"{test_case_id}/results/9876"
    intent_id = f"{agent_id}/intents/12345678"
    flows_map = {flow_id: "MOCK FLOW"}
    pages_map = {
        flow_id: {page_id: "MOCK PAGE"},
        other_flow_id: {other_page_id: "OTHER PAGE"},
    }

    return {
        "project_id": project_id,
        "agent_id": agent_id,
        "flow_id": flow_id,
        "page_id": page_id,
        "other_flow_id": other_flow_id,
        "other_page_id": other_page_id,
        "test_case_id": test_case_id,
        "test_case_result_id": test_case_result_id,
        "intent_id": intent_id,
        "flows_map": flows_map,
        "pages_map": pages_map,
    }


@pytest.fixture
def mock_tc_obj_turns(test_config):
    return types.TestCase(
        name=test_config["test_case_id"],
        tags=["#required", "#testing"],
        display_name="mock test case object",
        notes="This is a note for the mock test case",
        test_config=types.TestConfig(flow=test_config["flow_id"]),
        test_case_conversation_turns=[
            types.ConversationTurn(
                user_input=types.ConversationTurn.UserInput(
                    input=types.QueryInput(
                        text=types.TextInput(text="Hello"), language_code="en"
                    ),
                    injected_parameters={"my_key": "my_value"},
                    is_webhook_enabled=False,
                    enable_sentiment_analysis=False,
                )
            ),
            types.ConversationTurn(
                virtual_agent_output=types.ConversationTurn.VirtualAgentOutput(
                    triggered_intent=types.Intent(
                        name=test_config["intent_id"]
                    ),
                    text_responses=[
                        types.ResponseMessage.Text(
                            text=["Hi, how can I help you today?"]
                        )
                    ],
                )
            ),
        ],
        last_test_result=types.TestCaseResult(
            name=test_config["test_case_result_id"],
            test_result=types.TestResult(value=0),
        ),
    )


@pytest.fixture
def mock_tc_obj_no_turns(test_config):
    return types.TestCase(
        name=test_config["test_case_id"],
        tags=["#required", "#testing"],
        display_name="mock test case object",
        notes="This is a note for the mock test case",
        test_config=types.TestConfig(flow=test_config["flow_id"]),
        last_test_result=types.TestCaseResult(
            name=test_config["test_case_result_id"],
            test_result=types.TestResult(value=0),
        ),
    )


@pytest.fixture
def mock_tc_df(test_config):
    data = {
        "display_name": ["Mock Test 1", "Mock Test 2", "Mock Test 3"],
        "id": [
            f"{test_config['agent_id']}/testCases/12345",
            f"{test_config['agent_id']}/testCases/54321",
            f"{test_config['agent_id']}/testCases/67890",
        ],
        "short_id": ["12345", "54321", "67890"],
        "tags": [["tag1", "tag2"], ["tag3", "tag4"], ["tag5", "tag6"]],
        "creation_time": ["2023-01-01", "2023-01-02", "2023-01-03"],
        "start_flow": ["Mock Flow", "Mock Flow", "Default Start Flow"],
        "start_page": ["Mock Page", "Mock Page", "START_PAGE"],
        "passed": [True, False, True],
        "test_time": ["10:00", "11:00", "12:00"],
    }

    return pd.DataFrame(data)


@pytest.fixture
def mock_updated_tc_obj(mock_tc_obj_turns):
    mock_tc_obj_turns.display_name = "mock test case object updated"

    return mock_tc_obj_turns


@pytest.fixture
def mock_list_tc_pager(mock_tc_obj_turns):
    return test_cases.pagers.ListTestCasesPager(
        test_cases.TestCasesClient.list_test_cases,
        types.test_case.ListTestCasesRequest(),
        types.test_case.ListTestCasesResponse(test_cases=[mock_tc_obj_turns]),
    )


@pytest.fixture
def mock_list_tc_pager_no_turns(mock_tc_obj_no_turns):
    return test_cases.pagers.ListTestCasesPager(
        test_cases.TestCasesClient.list_test_cases,
        types.test_case.ListTestCasesRequest(),
        types.test_case.ListTestCasesResponse(
            test_cases=[mock_tc_obj_no_turns]
        ),
    )

@pytest.fixture(autouse=True)
def mock_client(test_config):
    """Fixture to create a mocked TestCasesClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.test_cases.services.test_cases.TestCasesClient") as mock_client:

        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client

def test_convert_test_result_to_string(mock_tc_obj_turns):
    tests = [(0, "TEST_RESULT_UNSPECIFIED"), (1, "PASSED"), (2, "FAILED")]
    tc = PyTestCases()

    for int_value, str_value in tests:
        mock_tc_obj_turns.last_test_result.test_result = types.TestResult(
            value=int_value
        )
        res = tc._convert_test_result_to_string(mock_tc_obj_turns)
        assert res == str_value
        assert isinstance(res, str)


def test_convert_test_result_to_bool(mock_tc_obj_turns):
    tests = [(0, None), (1, True), (2, False)]
    tc = PyTestCases()
    for int_value, res_value in tests:
        if int_value in [1, 2]:
            mock_tc_obj_turns.last_test_result.test_result = types.TestResult(
                value=int_value
            )
            res = tc._convert_test_result_to_bool(mock_tc_obj_turns)
            assert res == res_value
            assert isinstance(res_value, bool)
        else:
            mock_tc_obj_turns.last_test_result.test_result = types.TestResult(
                value=int_value
            )
            res = tc._convert_test_result_to_bool(mock_tc_obj_turns)
            assert res == res_value
            assert res_value is None


def test_get_flow_id_from_test_config(mock_tc_obj_turns, test_config):
    tests = [
        ("flow", test_config["flow_id"]),
        ("page", test_config["page_id"]),
        (None, ""),
    ]
    tc = PyTestCases()
    for str_value, test_case_id in tests:
        if str_value == "flow":
            mock_tc_obj_turns.test_config.flow = test_case_id
        elif str_value == "page":
            mock_tc_obj_turns.test_config.page = test_case_id
        else:
            mock_tc_obj_turns.test_config = types.TestConfig()

        res = tc._get_flow_id_from_test_config(mock_tc_obj_turns)
        assert isinstance(res, str)
        assert res == test_config["flow_id"]


def test_get_page_id_from_test_config(mock_tc_obj_turns, test_config):
    tests = [
        ("page", test_config["page_id"]),
        ("start_page", f"{test_config['flow_id']}/pages/START_PAGE"),
    ]
    tc = PyTestCases()
    for str_value, page_id in tests:
        if str_value == "page":
            mock_tc_obj_turns.test_config.page = page_id
            res = tc._get_page_id_from_test_config(
                mock_tc_obj_turns, test_config["flow_id"]
            )
            assert isinstance(str_value, str)
            assert res == page_id

        elif str_value == "start_page":
            mock_tc_obj_turns.test_config = types.TestConfig()
            res = tc._get_page_id_from_test_config(
                mock_tc_obj_turns, test_config["flow_id"]
            )
            assert isinstance(str_value, str)
            assert res == page_id


def test_get_page_display_name_flow_exists(test_config):
    tc = PyTestCases()

    res = tc._get_page_display_name(
        test_config["flow_id"], test_config["page_id"], test_config["pages_map"]
    )
    assert isinstance(res, str)
    assert res == "MOCK PAGE"


def test_get_page_display_name_flow_exists_no_page_map(test_config):
    tc = PyTestCases()

    res = tc._get_page_display_name(
        test_config["other_flow_id"], None, test_config["pages_map"]
    )
    assert res is None


def test_get_page_display_name_flow_does_not_exist(test_config):
    tc = PyTestCases()

    res = tc._get_page_display_name(
        f"{test_config['agent_id']}/flows/doesnt-exist-1234",
        None,
        test_config["pages_map"],
    )
    assert isinstance(res, str)
    assert res == "START_PAGE"


def test_process_test_case(mock_tc_obj_turns, test_config):
    cols = [
        "display_name",
        "id",
        "short_id",
        "tags",
        "creation_time",
        "start_flow",
        "start_page",
        "passed",
        "test_time",
    ]
    tc = PyTestCases()

    res = tc._process_test_case(
        mock_tc_obj_turns, test_config["flows_map"], test_config["pages_map"]
    )
    assert isinstance(res, pd.DataFrame)
    assert set(res.columns.to_list()) == set(cols)


@patch("dfcx_scrapi.core.test_cases.TestCases.batch_run_test_cases")
def test_retest_cases(mock_batch_run, mock_tc_df):
    cols = [
        "display_name",
        "id",
        "short_id",
        "tags",
        "creation_time",
        "start_flow",
        "start_page",
        "passed",
        "test_time",
    ]

    mock_batch_run.return_value.batch_run_test_cases.return_value = mock_tc_df

    retest_ids = mock_tc_df.id.to_list()[:1]
    tc = PyTestCases()
    res = tc._retest_cases(mock_tc_df, retest_ids)
    assert isinstance(res, pd.DataFrame)
    assert set(res.columns.to_list()) == set(cols)


# List Test Cases
def test_list_test_cases_agent_id_not_in_instance(
        mock_client, mock_tc_obj_turns):
    mock_client.return_value.list_test_cases.return_value = [mock_tc_obj_turns]

    tc = PyTestCases()

    with pytest.raises(ValueError):
        _ = tc.list_test_cases()


def test_list_test_cases_agent_id_in_instance(
    mock_client, mock_list_tc_pager, test_config
):
    mock_client.return_value.list_test_cases.return_value = mock_list_tc_pager

    tc = PyTestCases(agent_id=test_config["agent_id"])
    res = tc.list_test_cases()

    assert isinstance(res, list)
    assert isinstance(res[0], types.TestCase)


def test_list_test_cases_agent_id_in_method(
    mock_client, mock_list_tc_pager_no_turns, test_config
):
    mock_client.return_value.list_test_cases.return_value = (
        mock_list_tc_pager_no_turns
    )

    tc = PyTestCases()
    res = tc.list_test_cases(test_config["agent_id"])

    assert isinstance(res, list)
    assert isinstance(res[0], types.TestCase)
    assert res[0].test_case_conversation_turns == ""


def test_list_test_cases_include_conversation_turns(
    mock_client, mock_list_tc_pager, test_config
):
    mock_client.return_value.list_test_cases.return_value = mock_list_tc_pager

    tc = PyTestCases()
    res = tc.list_test_cases(
        test_config["agent_id"], include_conversation_turns=True
    )

    assert isinstance(res, list)
    assert isinstance(res[0], types.TestCase)
    assert res[0].test_case_conversation_turns != ""

def test_update_test_case_no_args(mock_client, mock_tc_obj_turns):
    mock_client.return_value.update_test_case.return_value = mock_tc_obj_turns

    tc = PyTestCases()

    # Assertions
    with pytest.raises(ValueError):
        _ = tc.update_test_case()

def test_update_test_case_kwargs_only(mock_client, mock_tc_obj_turns):
    mock_client.return_value.update_test_case.return_value = mock_tc_obj_turns

    tc = PyTestCases()

    # Assertions
    with pytest.raises(ValueError):
        _ = tc.update_test_case(display_name="mock test case object update")

def test_update_test_case_id_only(mock_client, test_config, mock_tc_obj_turns):

    mock_client.return_value.update_test_case.return_value = mock_tc_obj_turns

    tc = PyTestCases()

    # Assertions
    with pytest.raises(ValueError):
        _ = tc.update_test_case(test_case_id=test_config["test_case_id"])

def test_update_test_case_obj_only(mock_client, mock_tc_obj_turns):
    mock_client.return_value.update_test_case.return_value = mock_tc_obj_turns

    tc = PyTestCases()
    result = tc.update_test_case(obj=mock_tc_obj_turns)

    # Assertions
    assert result.display_name == "mock test case object"
    assert result == mock_tc_obj_turns

def test_update_test_case_obj_only_empty_name(mock_client, mock_tc_obj_turns):
    mock_tc_obj_turns.name = ""

    mock_client.return_value.update_test_case.return_value = mock_tc_obj_turns

    tc = PyTestCases()

    # Assertions
    with pytest.raises(ValueError):
        _ = tc.update_test_case(obj=mock_tc_obj_turns)

def test_update_test_case_with_obj_and_kwargs(
    mock_client, mock_tc_obj_turns, mock_updated_tc_obj
):

    mock_client.return_value.update_test_case.return_value = mock_updated_tc_obj

    tc = PyTestCases()
    result = tc.update_test_case(
        obj=mock_tc_obj_turns, display_name="mock test case object updated"
    )

    # Assertions
    assert result.display_name == mock_updated_tc_obj.display_name
    assert result == mock_updated_tc_obj

def test_update_test_case_with_id_and_kwargs(
    mock_client, test_config, mock_tc_obj_turns
):

    tc = PyTestCases()

    mock_client.return_value.get_test_case.return_value = mock_tc_obj_turns
    mock_client.return_value.update_test_case.return_value = mock_tc_obj_turns

    result = tc.update_test_case(
        test_case_id=test_config["test_case_id"],
        display_name="mock test case object update",
    )

    # Assertions
    assert result.display_name == "mock test case object update"

def test_update_test_case_with_obj_id_and_kwargs(
    mock_client, test_config, mock_tc_obj_turns
):
    """For this case, kwargs should be ignored in favor of the obj."""

    tc = PyTestCases()

    mock_client.return_value.get_test_case.return_value = mock_tc_obj_turns
    mock_client.return_value.update_test_case.return_value = mock_tc_obj_turns

    result = tc.update_test_case(
        test_case_id=test_config["test_case_id"],
        obj=mock_tc_obj_turns,
        display_name="mock test case object update",
    )

    # Assertions
    assert result.display_name == "mock test case object"
    assert result == mock_tc_obj_turns



"""Playbooks class for Generative Agents."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List

from google.cloud.dialogflowcx_v3beta1 import services, types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Playbooks(scrapi_base.ScrapiBase):
    """Core Class for CX Playbooks Resource functions."""

    def __init__(
        self,
        agent_id: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        playbooks_map: Dict[str, str] = None
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.agent_id = agent_id

        client_options = self._set_region(self.agent_id)
        self.playbooks_client = services.playbooks.PlaybooksClient(
            credentials=self.creds, client_options=client_options
        )
        self.agents_client = services.agents.AgentsClient(
            credentials=self.creds, client_options=client_options
        )

        self.playbooks_map = playbooks_map

    @staticmethod
    def build_instructions_from_list(
        instructions: List[str]) -> List[types.Playbook.Step]:
        """Helper method to create the playbook instruction set protos."""
        final_instructions = types.Playbook.Instruction(steps=[])

        for instruction in instructions:
            final_instructions.steps.append(
                types.Playbook.Step(text=instruction)
                )

        return final_instructions

    @staticmethod
    def clean_line(step_text: str):
        """Helper method to clean the current line that is being parsed."""
        step_text = step_text.strip() # clear any whitespace
        step_text = step_text.strip("-") # clear `-` if exists
        step_text = step_text.strip() # clear remaining whitespace if exists

        return step_text

    def parse_steps(
            self, lines: List[str], start_index: int, indent_level: int):
        """Recursively parse instructions and build Playbook.Step objects.

        Args:
            lines: The list of instruction lines.
            start_index: The index to start parsing from.
            indent_level: The current indentation level.

        Returns:
            A tuple containing:
                - A list of parsed Playbook.Step objects at the current level.
                - The index of the next line to parse.
        """
        steps: List[types.Playbook.Step] = []
        i = start_index
        while i < len(lines):
            line = lines[i]
            current_indent = len(line) - len(line.lstrip())
            if current_indent == indent_level:
                step_text = self.clean_line(line)
                step = types.Playbook.Step(text=step_text)
                steps.append(step)
                i += 1
            elif current_indent > indent_level:
                # Recursively parse child steps
                child_steps, next_index = self.parse_steps(
                    lines, i, current_indent)
                steps[-1].steps.extend(child_steps)
                i = next_index
            else:
                # Reached a line with lower indentation
                # stop parsing at this level
                break

        return steps, i

    def build_instructions_from_string(
            self,
            instructions: str) -> types.Playbook.Instruction:

        instruction_obj = types.Playbook.Instruction()
        lines = instructions.strip().splitlines()
        parsed_steps, _ = self.parse_steps(lines, 0, 0)
        instruction_obj.steps.extend(parsed_steps)

        return instruction_obj

    def process_playbook_kwargs(
            self,
            playbook: types.Playbook,
            **kwargs):
        """Process incoming kwargs and create the proper update mask."""
        paths = []
        for key, value in kwargs.items():
            if key in ["instruction", "instructions"]:
                if isinstance(value, list):
                    instructions = self.build_instructions_from_list(value)
                    setattr(playbook, "instruction", instructions)
                elif isinstance(value, str):
                    instructions = self.build_instructions_from_string(value)
                    setattr(playbook, "instruction", instructions)
                elif isinstance(value, types.Playbook.Instruction):
                    setattr(playbook, "instruction", value)
                paths.append("instruction")
            else:
                setattr(playbook, key, value)
                paths.append(key)

        # paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        return playbook, mask

    def set_default_playbook(self, playbook_id: str):
        """Sets the default Playbook for the Agent."""
        # Get the current Agent object and assign the new start_playbook ID
        agent = self.agents_client.get_agent(
            request=types.GetAgentRequest(
                name=self.agent_id
            )
        )
        agent.start_playbook = playbook_id

        # set the field mask for updating and update
        paths = {"start_playbook": playbook_id}.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        response = self.agents_client.update_agent(
            agent=agent, update_mask=mask)

        return response

    def get_playbooks_map(self, agent_id: str, reverse=False):
        """Exports Agent Playbook Names and UUIDs into a user friendly dict.

        Args:
          agent_id: the formatted CX Agent ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing Playbook UUIDs as keys and display names as
          values. If Optional reverse=True, the output will return playbook
          name:ID mapping instead of ID:playbook_name
        """
        if reverse:
            playbooks_dict = {
                playbook.display_name: playbook.name
                for playbook in self.list_playbooks(agent_id)
            }

        else:
            playbooks_dict = {
                playbook.name: playbook.display_name
                for playbook in self.list_playbooks(agent_id)
            }

        return playbooks_dict

    @scrapi_base.api_call_counter_decorator
    def list_playbooks(self, agent_id: str = None):
        """Get a List of all Playbooks in the current Agent.

        Args:
          agent_id: CX Agent ID string in the proper format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>

        Returns:
          List of Playbook objects
        """
        if not agent_id:
            agent_id = self.agent_id

        request = types.playbook.ListPlaybooksRequest()
        request.parent = agent_id

        response = self.playbooks_client.list_playbooks(request)

        playbooks = []
        for page in response.pages:
            for playbook in page.playbooks:
                playbooks.append(playbook)
        return playbooks

    @scrapi_base.api_call_counter_decorator
    def get_playbook(self, playbook_id: str = None):
        """Get a single CX Playbook object.

        Args:
          playbook_id: CX Playbook ID in the proper format

        Returns:
          A single CX Playbook object
        """
        if not playbook_id:
            playbook_id = self.playbook_id

        response = self.playbooks_client.get_playbook(name=playbook_id)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_playbook(
        self,
        agent_id: str,
        obj: types.Playbook = None,
        **kwargs,
    ):
        """Create a Dialogflow CX Playbook with given display name.

        If the user provides an existing Playbook object, a new CX Playbook
        will be created based on this object and any other input/kwargs will be
        discarded.

        Args:
          agent_id: DFCX Agent id where the Playbook will be created
          obj: (Optional) Playbook object to create in proto format

        Returns:
          The newly created CX Playbook resource object.
        """
        request = types.playbook.CreatePlaybookRequest()
        request.parent = agent_id

        if obj:
            playbook = obj
            playbook.name = ""
        else:
            playbook = types.Playbook()

            # set optional args as playbook attributes
            playbook, _ = self.process_playbook_kwargs(playbook, **kwargs)

        request.playbook = playbook
        response = self.playbooks_client.create_playbook(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def update_playbook(
        self,
        playbook_id: str,
        obj: types.Playbook = None,
        **kwargs
    ) -> types.Playbook:
        """Update a single specific CX Playbook object.

        Args:
          playbook_id: CX Playbook ID in the proper format
          obj: (Optional) a single CX Playbook object of types.Playbook
          overwrite_instructions: if True this will overwrite all instructions
            for the specific playbook. By default this is set to False and will
            append new instructions.

        Returns:
          A copy of the updated Playbook object
        """

        if obj:
            playbook = obj
            playbook.name = playbook_id
        else:
            playbook = self.get_playbook(playbook_id)

        # set optional args as playbook attributes
        playbook, mask = self.process_playbook_kwargs(playbook, **kwargs)

        response = self.playbooks_client.update_playbook(
            playbook=playbook, update_mask=mask)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_playbook(
        self, playbook_id: str = None, obj: types.Playbook = None):
        """Deletes a single CX Playbook Object resource.

        Args:
          playbook_id: The formatted CX Playbook ID to delete.
          obj: (Optional) a CX Playbook object of types.Playbook
        """
        if not playbook_id:
            playbook_id = self.playbook_id

        if obj:
            playbook_id = obj.name

        self.playbooks_client.delete_playbook(name=playbook_id)

    @scrapi_base.api_call_counter_decorator
    def create_playbook_version(
        self, playbook_id: str, description: str = None
        ) -> types.PlaybookVersion:
        """Creates a Playbook Version of the specific Playbook ID."""
        playbook_version = types.PlaybookVersion()
        playbook_version.name = playbook_id
        playbook_version.description = description

        request = types.CreatePlaybookVersionRequest()
        request.parent = playbook_id
        request.playbook_version = playbook_version

        response = self.playbooks_client.create_playbook_version(request)

        return response



""" Shared module to do deployement acting as a wrapper for deployment"""

import datetime
import json
import logging
import sys
import time

from google.cloud.dialogflowcx_v3beta1 import types

from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.environments import Environments
from dfcx_scrapi.core.flows import Flows
from dfcx_scrapi.core.versions import Versions

from .en_vs_other_lang import en_vs_lang
from .test_case_run import RunTestCases
from .webhook_update import update_webhook

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

class Deployment:
    """
    Manages the deployment and lifecycle of Dialogflow CX agents.

    This class provides methods for importing, testing, versioning, and
    deploying Dialogflow CX agents across different environments. It handles
    tasks such as:

    - Importing agents from GCS.
    - Updating webhook configurations.
    - Running test cases and validating results.
    - Collecting and managing flow IDs.
    - Versioning and deploying flows.
    - Syncing flows between environments (e.g., dev and prod).
    - Updating datastore settings.

    Attributes:
        (Initialized from an input dictionary)

    Methods:
        import_agent: Imports an agent from GCS to a target project.
        test_case_validation: Runs test cases and validates the results.
        collect_flow_id: Collects the IDs of flows to be deployed.
        version_count_delete: Manages version count and deletes old versions if
            necessary.
        version_cut: Creates new versions of the specified flows.
        deploy_versions: Deploys the new versions to the target environment.
        dev_prod_sync: Synchronizes flows between development and production
            environments.
        datastore_update: Updates datastore settings for the agent.
    """
    def __init__(self,input_dict):
        for key, value in input_dict.items():
            setattr(self, key, value)

    def import_agent(self,webhookenv):
        """Imports a Dialogflow CX agent to the target project.

        This method restores a Dialogflow CX agent from a GCS bucket to the
        specified target project and updates the webhook URI for the agent.

        Args:
            webhookenv: The webhook environment to use for the imported agent.
        """

        agent=Agents()
        target_agent_details=agent.get_agent_by_display_name(
            display_name=self.agent_name,
            project_id=self.target_project_id
        )

        self.target_agent_id=target_agent_details.name


        #restoring the agent from the SHA ID folder
        agent.restore_agent(
            agent_id=self.target_agent_id,
            gcs_bucket_uri=f"{self.sha_agent_gcs_location}/{self.agent_name}",
            restore_option=2
        )

        logging.info("import to destination project done")

        #[1.1] update webhooks uri
        update_webhook(self.target_agent_id,webhookenv)


    def test_case_validation(self):
        """Runs test cases and validates the results.

        This method executes test cases for the specified agent and environment,
        using tags to filter the test cases to run. If any test case fails,
        the script exits with an error code.

        Raises:
            SystemExit: If any test case fails.
        """

        tags=["#"+f for f in self.source_flow_names]
        obj=RunTestCases(
            project_id=self.target_project_id,
            agent_name=self.agent_name,
            environment_name=None)
        stats,result=obj.trigger_test_case(tags=tags)
        logging.info("test case result: %s", json.dumps(stats, indent=2))
        if not result:
            sys.exit(2)


    def collect_flow_id(self):
        """Collects the IDs of flows to be deployed.

        This method retrieves the IDs of the flows specified in
        `self.source_flow_names` from the target Dialogflow CX agent. It
        introduces a 50-second delay to allow for agent stabilization before
        fetching the flow IDs.
        """
        time.sleep(50)
        flow=Flows()
        logging.info(
            "flows to deployed in %s project: %s",
            self.target_project_id,
            self.source_flow_names
        )
        flow_ids=[]
        for flow_name in self.source_flow_names:
            flow_details=flow.get_flow_by_display_name(
                display_name=flow_name,
                agent_id=self.target_agent_id)
            flow_ids.append(flow_details.name
            )
        self.flow_ids=flow_ids

    def version_count_delete(self):
        """
        1. Check if the count of versions of a flow is not exceeding 20(limit)
        else delete the older version
        2. and make room for new version cut
        """
        versions=Versions()
        for flow_id in self.flow_ids:
            flowver=versions.list_versions(flow_id=flow_id)
        if len(flowver)==20:
            deletever=flowver[-1].name
            versions.delete_version(version_id=deletever)
            logging.info(
                "deleted version id %s in project %s",
                deletever,
                self.target_project_id
            )

    def version_cut(self):
        """
        1. Cut a version of those flows
        2. Storing new version ids created
        """
        versions=Versions()
        vers=[]
        for flow_id in self.flow_ids:
            v_display_name=f"version cut by CI/CD {datetime.datetime.now()}"
            ver=versions.create_version(
                flow_id=flow_id,
                description=self.updated_commit_message,
                display_name=v_display_name
            )
            vers.append(ver)

        #storing new version ids created
        new_versions=[]
        for ver in vers:
            verresult=ver.result()
            versionid=verresult.name
            new_versions.append(versionid)
        self.new_versions=new_versions
        logging.info("versions cut in %s project",self.target_project_id)

    def deploy_versions(self):
        """
        1.Deploy created versions to the env
        2.Deploy the version created to this env id
        """
        env=Environments()
        # get env id
        env_details=env.get_environment_by_display_name(
            display_name=self.target_environment_name,
            agent_id=self.target_agent_id
        )
        self.target_env_id=env_details.name

        # deploy the version created to this env id

        for new_version in self.new_versions:
            env.deploy_flow_to_environment(
                environment_id=self.target_env_id,
                flow_version=new_version)

        logging.info("versions deployed to deployed env %s project",
                     self.target_project_id
        )

    def dev_prod_sync(self):
        """
        sync the dev and prod project once deployment happens in prod
        1. Deploy created versions to the env
        2. Deploy the version created to this env id
        """
        agent=Agents()
        dev_agent_details=agent.get_agent_by_display_name(
            display_name=self.agent_name,
            project_id=self.dev_project
        )

        dev_agent_id=dev_agent_details.name
        env=Environments()
        # get env id
        env_details=env.get_environment_by_display_name(
            display_name=self.devprodsyncenv,
            agent_id=dev_agent_id
        )
        self.devprod_env_id=env_details.name

        # deploy the version created to this env id

        for new_version in self.impacted_version_ids:
            env.deploy_flow_to_environment(
                environment_id=self.devprod_env_id,
                flow_version=new_version)

        logging.info("flows deployed in prod is synced with dev environment")

    def datastore_update(self,projectlevel):
        """
        update the datastore id
        """
        if projectlevel=="uat":
            engine_id=self.uat_engine_id
        elif projectlevel=="uat":
            engine_id=self.prod_engine_id
        else:
            engine_id=""
        agents=Agents()
        app=types.Agent.GenAppBuilderSettings(engine=engine_id)
        kwargs={"gen_app_builder_settings":app}
        agents.update_agent(agent_id=self.target_agent_id,**kwargs)

        logging.info("datastore id updated")

    def fullfillment_lang_check(self,lang):
        """Checks fulfillment language coverage compared to English.

        This method compares the fulfillment coverage of the specified language
        (`lang`) with the English language ('en') for the given agent and flows.
        It returns dataframes containing statistics on fulfillment entries,
        parameters, and routes, along with a boolean result indicating whether
        all elements have agent responses in the specified language.

        Args:
            lang: The language code to compare against English (e.g., 'fr-ca').

        Returns:
            A tuple containing:
            - entry_df: DataFrame with statistics on entry fulfillment coverage.
            - param_df: DataFrame with statistics on parameter fulfillment
                coverage.
            - route_df: DataFrame with statistics on route fulfillment coverage.
            - result: A boolean indicating if all elements have agent responses
                in the specified language.
        """

        entry_df,param_df,route_df,result= en_vs_lang(
            self.target_agent_id,
            self.source_flow_names,
            lang
        )
        return entry_df,param_df,route_df,result




""" Running test cases and produce results"""

import logging
from typing import Dict, Tuple

from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.environments import Environments
from dfcx_scrapi.core.test_cases import TestCases

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

class RunTestCases:
    """
    Manages and executes test cases for Dialogflow CX agents.

    This class provides functionality to run test cases against a specified
    Dialogflow CX agent and environment. It retrieves the necessary agent and
    environment information and allows triggering test cases with optional tag
    filtering.

    Attributes:
        project_id: The ID of the Google Cloud project where the agent resides.
        agent_name: The display name of the agent.
        environment_name: The display name of the agent's environment
        (can be None).

    Methods:
        triggerTestcase: Executes test cases for the agent, optionally filtered
        by tags.
    """

    def __init__(
        self,project_id,
        agent_name,
        environment_name
    ):

        self.project_id=project_id
        self.agent_name=agent_name
        self.environment_name=environment_name

        agents=Agents()
        env=Environments()

        agent_details=agents.get_agent_by_display_name(
            display_name=self.agent_name,
            project_id=self.project_id)

        self.agent_id=agent_details.name

        #get environment id
        if self.environment_name:
            env_details=env.get_environment_by_display_name(
                display_name=self.environment_name,
                agent_id=self.agent_id)
            self.env_id=env_details.name
        else:
            self.env_id=None

    def trigger_test_case(
        self,
        tags,
        agent_id=None,
        env_id=None) -> Tuple[Dict[str, int], bool] :
        """
        Function to trigger the test case module in dfcx
        Returns:
        Result: Dict of results
        boolean mentioning test case status
        """
        if not agent_id:
            agent_id=self.agent_id
        if not env_id:
            env_id=self.env_id
        tc=TestCases()
        tc_list=tc.list_test_cases(agent_id=agent_id)

        #get test cases
        try:
            filtered_tc = [
                testcase
                for testcase in tc_list
                if any(
                    tag in testcase
                    for tag in tags
                )
            ]

        except AttributeError as e:
            print(
                f"Test case not found to run due to error {e}. "
                "UAT deployment will be done without test case validation"
            )
            result={"Pass": 0, "Fail": 0}
            return result, True
        filtered_tc_id=[filtestcase.name for filtestcase in filtered_tc]
        print(filtered_tc_id)

        #run the test cases
        tc_result=tc.batch_run_test_cases(test_cases=filtered_tc_id,
                        agent_id=agent_id,
                        environment=env_id)
        print(f"test case results {tc_result}")

        pass_count=0
        fail_count=0
        for result in tc_result.results:
            if result.test_result==1:
                pass_count+=1
            else:
                fail_count+=1

        print(f"Pass: {pass_count}, Fail: {fail_count}")
        result={"Pass": pass_count, "Fail": fail_count}

        if fail_count>0:
            return result,False
        return result,True



"""Test Class for FulfillmentBuilder in SCRAPI's builder package."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from dfcx_scrapi.builders.fulfillments import Fulfillment, FulfillmentBuilder


def test_create_new_proto_obj():
    fb = FulfillmentBuilder()
    assert fb.proto_obj is None

    fb.create_new_proto_obj()
    assert isinstance(fb.proto_obj, Fulfillment)
    assert fb.proto_obj.return_partial_responses is False
    assert fb.proto_obj.messages == []

    with pytest.raises(UserWarning):
        fb.create_new_proto_obj(return_partial_responses=True)

    fb.create_new_proto_obj(return_partial_responses=True, overwrite=True)
    assert fb.proto_obj.return_partial_responses is True


def test_create_new_proto_obj_with_webhook():
    fb = FulfillmentBuilder()
    custom_wbhk = (
        "projects/sample_project_id/locations/sample_location_id"
        "/agents/sample_agent_id/webhooks/sample_webhook_id")
    fb.create_new_proto_obj(
        webhook=custom_wbhk,
        tag="new_tag"
    )
    assert fb.proto_obj.webhook == custom_wbhk
    assert fb.proto_obj.tag == "new_tag"

    with pytest.raises(ValueError):
        fb.create_new_proto_obj(webhook=custom_wbhk, overwrite=True)


def test_add_parameter_presets_with_valid_dict():
    valid_param_map = {"p1": "v1", "p2": 123, "p3": True, "p4": None}
    fb = FulfillmentBuilder()
    fb.create_new_proto_obj()
    fb.add_parameter_presets(valid_param_map)

    for p in fb.proto_obj.set_parameter_actions:
        assert p.parameter in valid_param_map
        assert p.value == valid_param_map[p.parameter]

    new_params_map = {"n1": 12, "n2": "v2", "n3": 1.2}
    fb.add_parameter_presets(new_params_map)
    all_params_map = {**valid_param_map, **new_params_map}
    for p in fb.proto_obj.set_parameter_actions:
        assert p.parameter in all_params_map
        assert p.value == all_params_map[p.parameter]


def test_add_parameter_presets_with_invalid_dict():
    invalid_param_map = {"p1": "v1", 123: "p2"}
    fb = FulfillmentBuilder()
    fb.create_new_proto_obj()
    with pytest.raises(ValueError):
        fb.add_parameter_presets(invalid_param_map)

    # passing a list instead of dict
    with pytest.raises(ValueError):
        fb.add_parameter_presets(list(invalid_param_map.keys()))


def test_remove_parameter_presets_single_param():
    fb = FulfillmentBuilder()
    fb.create_new_proto_obj()
    params_map = {"p1": "v1", "p2": 123, "p3": True, "p4": None}
    fb.add_parameter_presets(parameter_map=params_map)
    fb.remove_parameter_presets(["p1"])
    params_map.pop("p1")
    for p in fb.proto_obj.set_parameter_actions:
        assert p.parameter in params_map
        assert p.value == params_map[p.parameter]


def test_remove_parameter_presets_multi_params():
    fb = FulfillmentBuilder()
    fb.create_new_proto_obj()
    params_map = {"p1": "v1", "p2": 123, "p3": True, "p4": None}
    fb.add_parameter_presets(parameter_map=params_map)
    fb.remove_parameter_presets(["p1", "p4"])
    params_map.pop("p1")
    params_map.pop("p4")
    for p in fb.proto_obj.set_parameter_actions:
        assert p.parameter in params_map
        assert p.value == params_map[p.parameter]


def test_has_webhook():
    fb = FulfillmentBuilder()
    fb.create_new_proto_obj()
    assert not fb.has_webhook()

    fb.create_new_proto_obj(
        webhook="sample_webhook_id", tag="some_tag", overwrite=True)
    assert fb.has_webhook()



"""Conversation History class for Generative Agents."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
from typing import Any, Dict, List

from google.cloud.dialogflowcx_v3beta1 import services, types
from tqdm.contrib.concurrent import thread_map

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class ConversationHistory(scrapi_base.ScrapiBase):
    """Used to get Conversation History Data."""
    def __init__(self,
                creds_path: str = None,
                creds_dict: Dict[str,str] = None,
                creds = None,
                scope = False,
                agent_id: str = None):
        super().__init__(creds_path, creds_dict, creds, scope)
        self.agent_id = agent_id

    @staticmethod
    def get_user_input(query_input: types.QueryInput):
        """Extract the input from the user."""
        if "text" in query_input:
            return query_input.text.text

        else:
            return None

    @staticmethod
    def get_query_result(query_result: types.QueryResult):
        """Extract the query result from the agent."""
        messages = []
        if "response_messages" in query_result:
            for rm in query_result.response_messages:
                if "text" in rm:
                    messages.append(rm.text.text[0])

        return " ".join(messages)

    def list_conversations(self, agent_id: str):
        request = types.conversation_history.ListConversationsRequest(
            parent=agent_id)

        client_options = self._set_region(agent_id)
        client = services.conversation_history.ConversationHistoryClient(
            credentials=self.creds, client_options=client_options)

        return list(client.list_conversations(request))

    def get_conversation(self, conversation_id: str):
        request = types.conversation_history.GetConversationRequest(
            name=conversation_id)
        client_options = self._set_region(conversation_id)
        client = services.conversation_history.ConversationHistoryClient(
            credentials=self.creds, client_options=client_options)

        return client.get_conversation(request)

    def delete_conversation(self, conversation_id: str) -> None:
        request = types.conversation_history.DeleteConversationRequest(
            name=conversation_id
        )
        client_options = self._set_region(conversation_id)
        client = services.conversation_history.ConversationHistoryClient(
            credentials=self.creds, client_options=client_options
        )

        client.delete_conversation(request)

    def process_single_conversation(self, current_convo: types.Conversation):
        """Extract details from single conversation for embed and cluster."""
        conversation = {}
        conversation["session_id"] = current_convo.name
        conversation["create_time"] = current_convo.start_time.rfc3339()
        conversation["turns"] = []
        for action in reversed(current_convo.interactions):
            turn = {}
            turn["user"] = self.get_user_input(action.request.query_input)
            turn["agent"] = self.get_query_result(action.response.query_result)
            conversation["turns"].append(turn)

        return conversation

    def write_conversations_to_file(
            self, convos: List[Dict[str, Any]], filename: str
            ):
        """Write conversations to file."""
        # Check if the file exists
        if not os.path.exists(filename):
            # If not, create an empty file
            with open(filename, "w", encoding="utf-8") as new_file:
                new_file.close()

        with open(filename, "w", encoding="utf-8") as json_file:
            for data_dict in convos:
                json.dump(data_dict, json_file)
                json_file.write("\n")

    def read_conversations_from_file(self, filename: str):
        """Loads a JSON Lines file and returns a list of dictionaries."""
        data = []
        with open(filename, "r", encoding="utf-8") as f:
            for line in f:
                data.append(json.loads(line))

        return data

    def conversation_history_to_file(self, agent_id: str, filename: str):
        """Process existing conversation history, with progress bar."""

        convo_ids = [convo.name for convo in self.list_conversations(agent_id)]

        def process_conversation(conversation_id: str):
            """Helper method to process single convo."""
            current_convo = self.get_conversation(conversation_id)
            return self.process_single_conversation(current_convo)

        # Use thread_map for progress visualization during parallel processing
        results = thread_map(
            process_conversation,
            convo_ids,
            desc="Processing Conversations"
            )

        self.write_conversations_to_file(list(results), filename)






"""Data Stores Resource functions for Vertex Search and Conversation."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List

from google.cloud.discoveryengine import (
    CreateDataStoreRequest,
    DataStore,
    DataStoreServiceClient,
    DeleteDataStoreRequest,
    GetDataStoreRequest,
    ListDataStoresRequest,
)
from google.longrunning.operations_pb2 import Operation

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class DataStores(scrapi_base.ScrapiBase):
    """Core Class for Data Store Resource functions."""
    def __init__(
        self,
        project_id: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.project_id = project_id

    @staticmethod
    def _get_content_config(content_type: str) -> DataStore.ContentConfig:
        """Build the Content Config used for the Data Store."""
        cc_map = {
            "unstructured": 1,
            "structured": 2,
            "website": 3
        }
        cc = DataStore.ContentConfig(cc_map[content_type])

        return cc

    def get_data_stores_map(self, location: str = "global",
                            reverse: bool = False) -> Dict[str,str]:
        """Get a user friendly mapping for Data Store Names and IDs."""
        if reverse:
            data_store_dict = {
                datastore.display_name: datastore.name
                for datastore in self.list_data_stores(location=location)
            }

        else:
            data_store_dict = {
                datastore.name: datastore.display_name
                for datastore in self.list_data_stores(location=location)
            }

        return data_store_dict

    def list_data_stores(
            self, location: str = "global") -> List[DataStore]:
        """List all data stores for a given project and location."""
        parent = self._build_data_store_parent(location)
        client_options = self._client_options_discovery_engine(
            f"projects/{self.project_id}/locations/{location}"
        )
        client = DataStoreServiceClient(
            credentials=self.creds, client_options=client_options
        )
        request = ListDataStoresRequest(parent=parent)
        page_result = client.list_data_stores(request=request)

        datastores = []
        for datastore in page_result:
            datastores.append(datastore)

        return datastores

    def get_data_store(self, data_store_id: str) -> DataStore:
        """Get a single Data Store by specified ID."""
        client_options = self._client_options_discovery_engine(data_store_id)
        client = DataStoreServiceClient(
            credentials=self.creds, client_options=client_options
        )
        request = GetDataStoreRequest(
            name=data_store_id
            )

        response = client.get_data_store(request)

        return response

    def create_datastore(
            self, display_name: str, solution_type: str = "chat",
            datastore_type: str = "website", advanced_site_search: bool = True,
            location: str = "global") -> Operation:
        """Create a new Data Store type specified by the user.

        Args:
          display_name: The data store display name. This field must be a UTF-8
            encoded string with a length limit of 128 characters. Otherwise, an
            INVALID_ARGUMENT error is returned.
          solution_type: ONEOF [`chat`, `search`, `recommendation`]
          datastore_type: ONEOF [`website`, `structured`, `unstructured`]
          advanced_site_search: A boolean flag indicating whether user wants to
            directly create an advanced data store for site search. If the data
            store is not configured as site search (GENERIC vertical and
            PUBLIC_WEBSITE content_config), this flag will be ignored.
          location: the GCP region to create the Data Store in
        """
        parent = self._build_data_store_parent(location)
        client_options = self._client_options_discovery_engine(
            f"projects/{self.project_id}/locations/{location}"
        )
        client = DataStoreServiceClient(
            credentials=self.creds, client_options=client_options
        )
        data_store = DataStore()
        data_store.display_name = display_name
        data_store.industry_vertical = 1
        data_store.solution_types = [self._get_solution_type(solution_type)]
        data_store.content_config = self._get_content_config(datastore_type)

        request = CreateDataStoreRequest(
            parent=parent,
            data_store=data_store,
            data_store_id=data_store.display_name,
            create_advanced_site_search=advanced_site_search
        )

        operation = client.create_data_store(request=request)

        return operation.operation

    def delete_datastore(self, data_store_id: str):
        """Delete the specified Data Store by ID."""
        client_options = self._client_options_discovery_engine(data_store_id)
        client = DataStoreServiceClient(
            credentials=self.creds, client_options=client_options
        )

        request = DeleteDataStoreRequest(
            name=data_store_id
        )
        operation = client.delete_data_store(request=request)

        return operation.operation



""" Getting impacted flow functions"""

from typing import Dict

from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.core.environments import Environments
from dfcx_scrapi.core.flows import Flows


class Impacted:
    """
    Analyzes and identifies changes in Dialogflow CX agent flows across
    environment versions.

    This class retrieves information about a specified Dialogflow CX agent and
    its environment, including version history and flow details. It then
    compares the latest two versions to identify any changes in the flows,
    providing a mapping of impacted flow IDs and names.

    Attributes:
        source_project_id: The ID of the Google Cloud project where the agent
            resides.
        source_agent_name: The display name of the agent.
        environment_name: The display name of the agent's environment
            (default: "ready to deploy").

    Methods:
        filter_flows: (Static method) Filters a flow map based on
        differences between two versions in the same environments.
        check_flow: Identifies and returns a dictionary of
        changed flows between the latest two versions.
    """

    # Get agent id

    def __init__(
        self, source_project_id,
        source_agent_name,
        environment_name="ready to deploy"
    ):
        self.env = Environments()
        self.flows = Flows()

        self.source_project_id = source_project_id
        self.source_agent_name = source_agent_name
        self.environment_name = environment_name
        self.filtered_dict = {}

        agents = Agents()
        agent_details = agents.get_agent_by_display_name(
            display_name=self.source_agent_name,
            project_id=self.source_project_id
        )

        self.agent_id = agent_details.name

        # get environment id
        env_details = self.env.get_environment_by_display_name(
            display_name=self.environment_name,
            agent_id=self.agent_id
        )
        self.env_id = env_details.name

        # get history
        self.hist = self.env.lookup_environment_history(
            environment_id=self.env_id
        )

    @staticmethod
    def filter_flows(env_ver1, env_ver2, flowmap, versions):
        """Returns filtered dict and impacted version ids"""
        impacted_flows = []
        for k, v in env_ver1.items():
            if v != env_ver2.get(k, 0):
                impacted_flows.append(k)

        filtered_dict = {
            k: v for k, v in flowmap.items()
            if k.split("/")[-1] in impacted_flows
        }
        # getting version ids
        impacted_version_ids = []
        for version_info in versions:
            ver_id_str = version_info.version
            flow = ver_id_str.split("/")[-3]
            if flow in impacted_flows:
                impacted_version_ids.append(ver_id_str)


        return filtered_dict, impacted_version_ids



    def check_flow(
        self
        ) -> Dict[str, str]:
        # compare latest 2 history
        """
        returns map of flow id:flow name which was found to be changed
        """
        env_ver1 = {}
        for i in self.hist[0].version_configs:
            flow = i.version.split("/")[-3]
            version = i.version.split("/")[-1]
            env_ver1[flow] = version

        env_ver2 = {}
        if len(self.hist) > 1:
            for i in self.hist[1].version_configs:
                flow = i.version.split("/")[-3]
                version = i.version.split("/")[-1]
                env_ver2[flow] = version

        # get flow map for id name comparision
        flowmap = self.flows.get_flows_map(agent_id=self.agent_id)

        self.filtered_dict, self.impacted_version_ids = Impacted.filter_flows(
            env_ver1,
            env_ver2,
            flowmap,
            self.hist[0].version_configs
        )

        return self.filtered_dict, self.impacted_version_ids



"""Unit Tests for Examples."""

# pylint: disable=redefined-outer-name
# pylint: disable=protected-access
# pylint: disable=missing-function-docstring

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import pytest
from google.cloud.dialogflowcx_v3beta1 import services, types

from dfcx_scrapi.core.examples import Examples


@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    playbook_id = f"{agent_id}/playbooks/1234"
    example_id = f"{playbook_id}/examples/9876"
    tool_id = f"{agent_id}/tools/4321"
    display_name = "test_example"
    return {
        "project_id": project_id,
        "agent_id": agent_id,
        "playbook_id": playbook_id,
        "example_id": example_id,
        "tool_id": tool_id,
        "display_name": display_name
        }

@pytest.fixture
def mock_example_obj(test_config):
    tool_use = types.ToolUse()
    tool_use.tool = test_config["tool_id"]
    tool_use.action = "places_search_tool"
    tool_use.input_action_parameters = {
        "preferences": "dim sum restaurant",
        "city": "Hoboken"
        }
    tool_use.output_action_parameters = {
    "results": [
        {
        "address": "512 Washington St, Hoboken, NJ 07030, United States",
        "name": "South Lions Dim Sum & Tea",
        "user_ratings_total": 141,
        },
        {
        "name": "Precious Chinese & Japanese Cuisine",
        "address": "128 Washington St, Hoboken, NJ 07030, United States",
        "user_ratings_total": 278,
        },
        {
        "name": "Keming",
        "user_ratings_total": 224,
        "address": "1006 Washington St, Hoboken, NJ 07030, United States"
        }
    ]
    }
    action = types.Action()
    action.tool_use = tool_use

    return types.Example(
        name=test_config["example_id"],
        display_name=test_config["display_name"],
        actions=[action]
        )

@pytest.fixture
def mock_updated_example_obj(mock_example_obj):
    mock_example_obj.display_name = "updated_test_example"
    return mock_example_obj

@pytest.fixture
def mock_list_examples_pager(mock_example_obj):
    return services.examples.pagers.ListExamplesPager(
        services.examples.ExamplesClient.list_examples,
        types.example.ListExamplesRequest(),
        types.example.ListExamplesResponse(examples=[mock_example_obj]),
        )

@pytest.fixture(autouse=True)
def mock_client(test_config):
    """Fixture to create mocked ExamplesClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.examples.services.examples.ExamplesClient") as mock_client, \
        patch("dfcx_scrapi.core.playbooks.Playbooks.__init__") as mock_playbooks_init, \
        patch("dfcx_scrapi.core.tools.Tools.__init__") as mock_tools_init:

        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()
        mock_playbooks_init.return_value = None
        mock_tools_init.return_value = None

        yield mock_client 

# Test get_examples_map
def test_get_examples_map(mock_client, mock_list_examples_pager, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    mock_client.return_value.list_examples.return_value = (
        mock_list_examples_pager
        )
    res = ex.get_examples_map(playbook_id=test_config["playbook_id"])

    assert isinstance(res, dict)
    assert test_config["example_id"] in res
    assert res[test_config["example_id"]] == test_config["display_name"]

    print(mock_client.mock_calls)

# Test list_examples
def test_list_examples(mock_client, mock_list_examples_pager, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    mock_client.return_value.list_examples.return_value = (
    mock_list_examples_pager
    )
    res = ex.list_examples(playbook_id=test_config["playbook_id"])

    assert isinstance(res, list)
    assert isinstance(res[0], types.Example)

# Test get_example
def test_get_example(mock_client, mock_example_obj, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    mock_client.return_value.get_example.return_value = mock_example_obj
    res = ex.get_example(example_id=test_config["example_id"])

    assert isinstance(res, types.Example)
    assert res.display_name == test_config["display_name"]

# Test create_example
def test_create_example_from_kwargs(
    mock_client, mock_example_obj, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    mock_client.return_value.create_example.return_value = mock_example_obj
    res = ex.create_example(
    playbook_id=test_config["playbook_id"],
    display_name=test_config["display_name"]
    )
    assert isinstance(res, types.Example)
    assert res.display_name == test_config["display_name"]

def test_create_example_from_proto_object(
    mock_client, mock_example_obj, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    mock_client.return_value.create_example.return_value = mock_example_obj
    res = ex.create_example(
    playbook_id=test_config["playbook_id"],
    obj=mock_example_obj
    )
    assert isinstance(res, types.Example)
    assert res.display_name == test_config["display_name"]

# Test update_example
def test_update_example_with_obj(
    mock_client, mock_updated_example_obj, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    mock_client.return_value.update_example.return_value = (
    mock_updated_example_obj
    )
    res = ex.update_example(
    example_id=test_config["example_id"],
    obj=mock_updated_example_obj
    )

    assert isinstance(res, types.Example)
    assert res.display_name == "updated_test_example"

def test_update_example_with_kwargs(
    mock_client, mock_example_obj, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    mock_client.return_value.get_example.return_value = mock_example_obj
    mock_client.return_value.update_example.return_value = mock_example_obj
    res = ex.update_example(
        example_id=test_config["example_id"],
        display_name="updated_test_example"
        )

    assert isinstance(res, types.Example)
    assert res.display_name == "updated_test_example"

# Test delete_example
def test_delete_example(mock_client, test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    ex.delete_example(example_id=test_config["example_id"])
    mock_client.return_value.delete_example.assert_called()

# Test get_playbook_state
def test_get_playbook_state(test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    assert ex.get_playbook_state("OK") == 1
    assert ex.get_playbook_state("CANCELLED") == 2
    assert ex.get_playbook_state("FAILED") == 3
    assert ex.get_playbook_state("ESCALATED") == 4
    assert ex.get_playbook_state("PENDING") == 5
    assert ex.get_playbook_state(None) == 0

# Test build_example_from_action_list_dict
def test_build_example_from_action_list(test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    action_list = [
    {"user_utterance": "hello"},
    {"agent_utterance": "hi there"},
    ]
    example = ex.build_example_from_action_list(
        display_name="test_example", action_list=action_list
        )
    assert isinstance(example, types.Example)
    assert example.display_name == "test_example"
    assert len(example.actions) == 2

# Test build_playbook_invocation
def test_build_playbook_invocation(test_config):
    ex = Examples(agent_id=test_config["agent_id"])
    ex.playbooks_map = {"test_playbook": test_config["playbook_id"]}

    action = {"playbook_name": "test_playbook"}
    pb_inv = ex.build_playbook_invocation(action)
    assert isinstance(pb_inv, types.PlaybookInvocation)
    assert pb_inv.playbook == test_config["playbook_id"]



"""Webhook Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict

from google.cloud.dialogflowcx_v3beta1 import services, types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Webhooks(scrapi_base.ScrapiBase):
    """Core Class for CX Webhook Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        webhook_id: str = None,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if webhook_id:
            self.webhook_id = webhook_id
            self.client_options = self._set_region(webhook_id)

        if agent_id:
            self.agent_id = agent_id

    def get_webhooks_map(
        self,
        agent_id: str = None,
        reverse=False):
        """Exports Agent Webhook Names and UUIDs into a user friendly dict.

        Args:
          agent_id: the formatted CX Agent ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing Webhook UUIDs as keys and
          webhook display names as values
        """
        if not agent_id:
            agent_id = self.agent_id

        if reverse:
            webhooks_dict = {
                webhook.display_name: webhook.name
                for webhook in self.list_webhooks(agent_id)
            }

        else:
            webhooks_dict = {
                webhook.name: webhook.display_name
                for webhook in self.list_webhooks(agent_id)
            }

        return webhooks_dict


    @scrapi_base.api_call_counter_decorator
    def list_webhooks(self, agent_id: str = None):
        """List all Webhooks in the specified CX Agent.

        Args:
          agent_id: the formated CX Agent ID to use

        Returns:
          List of webhook objects
        """
        if not agent_id:
            agent_id = self.agent_id

        request = types.webhook.ListWebhooksRequest()
        request.parent = agent_id

        client_options = self._set_region(agent_id)
        client = services.webhooks.WebhooksClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_webhooks(request)

        cx_webhooks = []
        for page in response.pages:
            for cx_webhook in page.webhooks:
                cx_webhooks.append(cx_webhook)

        return cx_webhooks

    @scrapi_base.api_call_counter_decorator
    def create_webhook(
        self,
        agent_id: str,
        obj: types.Webhook = None,
        **kwargs):
        """Create a single webhook resource on a given CX Agent.

        Args:
          agent_id: the formatted CX Agent ID to create the webhook on
          obj: (Optional) the Webhook object of type
            types.Webhook that you want to create the webhook from

        Returns:
          The successfully created webhook object
        """
        if obj:
            webhook = obj
            webhook.name = ""
        else:
            webhook = types.webhook.Webhook()

        # set optional kwargs to webhook attributes
        for key, value in kwargs.items():
            setattr(webhook, key, value)

        client_options = self._set_region(agent_id)
        client = services.webhooks.WebhooksClient(
            client_options=client_options, credentials=self.creds)
        response = client.create_webhook(parent=agent_id, webhook=webhook)

        return response


    @scrapi_base.api_call_counter_decorator
    def get_webhook(self, webhook_id:str):
        """Retrieves the specified webhook.

        Args:
          webhook_id: The ID of the webhook. Format:
            projects/<Project ID>/locations/<Location ID>/agents/
            <Agent ID>/webhooks/<Webhook

        Returns:
          The Webhook object.
        """
        request = types.webhook.GetWebhookRequest()
        request.name = webhook_id

        client_options = self._set_region(webhook_id)
        client = services.webhooks.WebhooksClient(
            client_options=client_options, credentials=self.creds)

        response = client.get_webhook(request)

        return response

    def get_webhook_by_display_name(
        self,
        webhook_display_name:str,
        agent_id:str = None):
        """Retrieves the specified webhook.

        Args:
          webhook_webhook_name: The display name of the webhook.
         agent_id: Optional. The formatted CX Agent ID.

        Returns:
          The Webhook object of the specified webhook.
        """

        if not agent_id:
            agent_id = self.agent_id

        webhook_map = self.get_webhooks_map(agent_id=agent_id,reverse=True)

        if webhook_display_name in webhook_map:
            webhook_obj = self.get_webhook(webhook_map[webhook_display_name])

        else:
            raise ValueError(
                f"Webhook \"{webhook_display_name}\" does not exist in the \
                    specified Agent."
                )

        return webhook_obj


    @scrapi_base.api_call_counter_decorator
    def update_webhook(
        self,
        webhook_id:str,
        webhook_obj:types.Webhook = None,
        **kwargs):
        """Update the values of an existing webhook.

        Args:
          webhook_id: The ID of the webhook. Format:
            projects/<Project ID>/locations/<Location ID>/agents/
            <Agent ID>/webhooks/<Webhook ID>
          webhook_obj: Optional Webhook object of types.Webhook
            that can be provided when you are planning to replace the full
            object vs. just partial updates.

        Returns:
          The Webhook object with specified changes.
        """
        if not webhook_obj:
            webhook_obj = types.Webhook()

        webhook_obj.name = webhook_id

        # set environment attributes from kwargs
        for key, value in kwargs.items():
            setattr(webhook_obj, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(webhook_id)
        client = services.webhooks.WebhooksClient(
            client_options=client_options, credentials=self.creds)

        request = types.webhook.UpdateWebhookRequest()
        request.webhook = webhook_obj
        request.update_mask = mask

        response = client.update_webhook(request)

        return response


    @scrapi_base.api_call_counter_decorator
    def delete_webhook(
        self, webhook_id: str = None,
        obj: types.Webhook = None, force: bool = False
    ):
        """Deletes a single Webhookd resource object.

        Args:
          webhook_id: intent to delete
          obj: (Optional) a CX Webhook object of types.Webhook
          force: (Optional) This field has no effect for webhook not being
            used. For webhooks that are used by pages/flows/transition route
            groups:
            -  If ``force`` is set to false, an error will be returned
               with message indicating the referenced resources.
            -  If ``force`` is set to true, Dialogflow will remove the
               webhook, as well as any references to the webhook and tags in
               fulfillments that point to this webhook will be removed.
        """
        if not webhook_id:
            webhook_id = self.webhook_id

        if obj:
            webhook_id = obj.name

        client_options = self._set_region(webhook_id)
        client = services.webhooks.WebhooksClient(
            client_options=client_options, credentials=self.creds)
        req = types.DeleteWebhookRequest(name=webhook_id, force=force)
        client.delete_webhook(request=req)
